#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import fnmatch

from PyInstaller import compat
from PyInstaller import isolated
from PyInstaller import log as logging
from PyInstaller.depend import bindepend

if compat.is_darwin:
    from PyInstaller.utils import osx as osxutils

logger = logging.getLogger(__name__)


@isolated.decorate
def _get_tcl_tk_info():
    """
    Isolated-subprocess helper to retrieve the basic Tcl/Tk information:
     - tkinter_extension_file = the value of __file__ attribute of the _tkinter binary extension (path to file).
     - tcl_data_dir = path to the Tcl library/data directory.
     - tcl_version = Tcl version
     - tk_version = Tk version
     - tcl_theaded = boolean indicating whether Tcl/Tk is built with multi-threading support.
    """
    try:
        import tkinter
        import _tkinter
    except ImportError:
        # tkinter unavailable
        return None
    try:
        tcl = tkinter.Tcl()
    except tkinter.TclError:  # e.g. "Can't find a usable init.tcl in the following directories: ..."
        return None

    # Query the location of Tcl library/data directory.
    tcl_data_dir = tcl.eval("info library")

    # Check if Tcl/Tk is built with multi-threaded support (built with --enable-threads), as indicated by the presence
    # of optional `threaded` member in `tcl_platform` array. Tcl 9.0 removed the --enable-threads flag, and is always
    # built with multi-threaded support (and thus the `threaded` array member has been removed).
    TCL_MAJOR = int(_tkinter.TCL_VERSION.split(".")[0])
    if TCL_MAJOR >= 9:
        tcl_threaded = True
    else:
        try:
            tcl.getvar("tcl_platform(threaded)")  # Ignore the actual value.
            tcl_threaded = True
        except tkinter.TclError:
            tcl_threaded = False

    return {
        "available": True,
        # If `_tkinter` is a built-in (as opposed to an extension), it does not have a `__file__` attribute.
        "tkinter_extension_file": getattr(_tkinter, '__file__', None),
        "tcl_version": _tkinter.TCL_VERSION,
        "tk_version": _tkinter.TK_VERSION,
        "tcl_threaded": tcl_threaded,
        "tcl_data_dir": tcl_data_dir,
    }


class TclTkInfo:
    # Root directory names of Tcl and Tk library/data directories in the frozen application. These directories are
    # originally fully versioned (e.g., tcl8.6 and tk8.6); we want to remap them to unversioned variants, so that our
    # run-time hook (pyi_rthook__tkinter.py) does not have to determine version numbers when setting `TCL_LIBRARY`
    # and `TK_LIBRARY` environment variables.
    #
    # We also cannot use plain "tk" and "tcl", because on macOS, the Tcl and Tk shared libraries might come from
    # framework bundles, and would therefore end up being collected as "Tcl" and "Tk" in the top-level application
    # directory, causing clash due to filesystem being case-insensitive by default.
    TCL_ROOTNAME = '_tcl_data'
    TK_ROOTNAME = '_tk_data'

    def __init__(self):
        pass

    def __repr__(self):
        return "TclTkInfo"

    # Delay initialization of Tcl/Tk information until until the corresponding attributes are first requested.
    def __getattr__(self, name):
        if 'available' in self.__dict__:
            # Initialization was already done, but requested attribute is not available.
            raise AttributeError(name)

        # Load Qt library info...
        self._load_tcl_tk_info()
        # ... and return the requested attribute
        return getattr(self, name)

    def _load_tcl_tk_info(self):
        logger.info("%s: initializing cached Tcl/Tk info...", self)

        # Initialize variables so that they might be accessed even if tkinter/Tcl/Tk is unavailable or if initialization
        # fails for some reason.
        self.available = False
        self.tkinter_extension_file = None
        self.tcl_version = None
        self.tk_version = None
        self.tcl_threaded = False
        self.tcl_data_dir = None

        self.tk_data_dir = None
        self.tcl_module_dir = None

        self.is_macos_system_framework = False
        self.tcl_shared_library = None
        self.tk_shared_library = None

        self.data_files = []

        try:
            tcl_tk_info = _get_tcl_tk_info()
        except Exception as e:
            logger.warning("%s: failed to obtain Tcl/Tk info: %s", self, e)
            return

        # If tkinter could not be imported, `_get_tcl_tk_info` returns None. In such cases, emit a debug message instead
        # of a warning, because this initialization might be triggered by a helper function that is trying to determine
        # availability of `tkinter` by inspecting the `available` attribute.
        if tcl_tk_info is None:
            logger.debug("%s: failed to obtain Tcl/Tk info: tkinter/_tkinter could not be imported.", self)
            return

        # Copy properties
        for key, value in tcl_tk_info.items():
            setattr(self, key, value)

        # Parse Tcl/Tk version into (major, minor) tuple.
        self.tcl_version = tuple((int(x) for x in self.tcl_version.split(".")[:2]))
        self.tk_version = tuple((int(x) for x in self.tk_version.split(".")[:2]))

        # Determine full path to Tcl and Tk shared libraries against which the `_tkinter` extension module is linked.
        # This can only be done when `_tkinter` is in fact an extension, and not a built-in. In the latter case, the
        # Tcl/Tk libraries are statically linked into python shared library, so there are no shared libraries for us
        # to discover.
        if self.tkinter_extension_file:
            try:
                (
                    self.tcl_shared_library,
                    self.tk_shared_library,
                ) = self._find_tcl_tk_shared_libraries(self.tkinter_extension_file)
            except Exception:
                logger.warning("%s: failed to determine Tcl and Tk shared library location!", self, exc_info=True)

            # macOS: check if _tkinter is linked against system-provided Tcl.framework and Tk.framework. This is the
            # case with python3 from XCode tools (and was the case with very old homebrew python builds). In such cases,
            # we should not be collecting Tcl/Tk files.
            if compat.is_darwin:
                self.is_macos_system_framework = self._check_macos_system_framework(self.tcl_shared_library)

                # Emit a warning in the unlikely event that we are dealing with Teapot-distributed version of ActiveTcl.
                if not self.is_macos_system_framework:
                    self._warn_if_using_activetcl_or_teapot(self.tcl_data_dir)

        # Infer location of Tk library/data directory. Ideally, we could infer this by running
        #
        # import tkinter
        # root = tkinter.Tk()
        # tk_data_dir = root.tk.exprstring('$tk_library')
        #
        # in the isolated subprocess as part of `_get_tcl_tk_info`. However, that is impractical, as it shows the empty
        # window, and on some platforms (e.g., linux) requires display server. Therefore, try to guess the location,
        # based on the following heuristic:
        #  - if TK_LIBRARY is defined use it.
        #  - if Tk is built as macOS framework bundle, look for Scripts sub-directory in Resources directory next to
        #    the shared library.
        #  - otherwise, look for: $tcl_root/../tkX.Y, where X and Y are Tk major and minor version.
        if "TK_LIBRARY" in os.environ:
            self.tk_data_dir = os.environ["TK_LIBRARY"]
        elif compat.is_darwin and self.tk_shared_library and (
            # is_framework_bundle_lib handles only fully-versioned framework library paths...
            (osxutils.is_framework_bundle_lib(self.tk_shared_library)) or
            # ... so manually handle top-level-symlinked variant for now.
            (self.tk_shared_library).endswith("Tk.framework/Tk")
        ):
            # Fully resolve the library path, in case it is a top-level symlink; for example, resolve
            # /Library/Frameworks/Python.framework/Versions/3.13/Frameworks/Tk.framework/Tk
            # into
            # /Library/Frameworks/Python.framework/Versions/3.13/Frameworks/Tk.framework/Versions/8.6/Tk
            tk_lib_realpath = os.path.realpath(self.tk_shared_library)
            # Resources/Scripts directory next to the shared library
            self.tk_data_dir = os.path.join(os.path.dirname(tk_lib_realpath), "Resources", "Scripts")
        else:
            self.tk_data_dir = os.path.join(
                os.path.dirname(self.tcl_data_dir),
                f"tk{self.tk_version[0]}.{self.tk_version[1]}",
            )

        # Infer location of Tcl module directory. The modules directory is separate from the library/data one, and
        # is located at $tcl_root/../tclX, where X is the major Tcl version.
        self.tcl_module_dir = os.path.join(
            os.path.dirname(self.tcl_data_dir),
            f"tcl{self.tcl_version[0]}",
        )

        # Find all data files
        if self.is_macos_system_framework:
            logger.info("%s: using macOS system Tcl/Tk framework - not collecting data files.", self)
        else:
            # Collect Tcl and Tk scripts from their corresponding library/data directories. See comment at the
            # definition of TK_ROOTNAME and TK_ROOTNAME variables.
            if os.path.isdir(self.tcl_data_dir):
                self.data_files += self._collect_files_from_directory(
                    self.tcl_data_dir,
                    prefix=self.TCL_ROOTNAME,
                    excludes=['demos', '*.lib', 'tclConfig.sh'],
                )
            else:
                logger.warning("%s: Tcl library/data directory %r does not exist!", self, self.tcl_data_dir)

            if os.path.isdir(self.tk_data_dir):
                self.data_files += self._collect_files_from_directory(
                    self.tk_data_dir,
                    prefix=self.TK_ROOTNAME,
                    excludes=['demos', '*.lib', 'tkConfig.sh'],
                )
            else:
                logger.warning("%s: Tk library/data directory %r does not exist!", self, self.tk_data_dir)

            # Collect Tcl modules from modules directory
            if os.path.isdir(self.tcl_module_dir):
                self.data_files += self._collect_files_from_directory(
                    self.tcl_module_dir,
                    prefix=os.path.basename(self.tcl_module_dir),
                )
            else:
                logger.warning("%s: Tcl module directory %r does not exist!", self, self.tcl_module_dir)

    @staticmethod
    def _collect_files_from_directory(root, prefix=None, excludes=None):
        """
        A minimal port of PyInstaller.building.datastruct.Tree() functionality, which allows us to avoid using Tree
        here. This way, the TclTkInfo data structure can be used without having PyInstaller's config context set up.
        """
        excludes = excludes or []

        todo = [(root, prefix)]
        output = []
        while todo:
            target_dir, prefix = todo.pop()

            for entry in os.listdir(target_dir):
                # Basic name-based exclusion
                if any((fnmatch.fnmatch(entry, exclude) for exclude in excludes)):
                    continue

                src_path = os.path.join(target_dir, entry)
                dest_path = os.path.join(prefix, entry) if prefix else entry

                if os.path.isdir(src_path):
                    todo.append((src_path, dest_path))
                else:
                    # Return 3-element tuples with fully-resolved dest path, since other parts of code depend on that.
                    output.append((dest_path, src_path, 'DATA'))

        return output

    @staticmethod
    def _find_tcl_tk_shared_libraries(tkinter_ext_file):
        """
        Find Tcl and Tk shared libraries against which the _tkinter extension module is linked.
        """
        tcl_lib = None
        tk_lib = None

        for _, lib_path in bindepend.get_imports(tkinter_ext_file):  # (name, fullpath) tuple
            if lib_path is None:
                continue  # Skip unresolved entries

            # For comparison, take basename of lib_path. On macOS, lib_name returned by get_imports is in fact
            # referenced name, which is not necessarily just a basename.
            lib_name = os.path.basename(lib_path)
            lib_name_lower = lib_name.lower()  # lower-case for comparisons

            # First check for Tk library, because it is unlikely that 'tk' will appear in the name of the Tcl shared
            # library, while 'tcl' could appear in the name of the Tk shared library. For example, Fedora 43 ships
            # both Tcl/Tk 8.6 and 9.0, and in the latter, the libraries are named `libtcl9.0.so` and `libtcl9tk9.0.so`.
            if 'tk' in lib_name_lower:
                tk_lib = lib_path
            elif 'tcl' in lib_name_lower:
                tcl_lib = lib_path

        return tcl_lib, tk_lib

    @staticmethod
    def _check_macos_system_framework(tcl_shared_lib):
        # Starting with macOS 11, system libraries are hidden (unless both Python and PyInstaller's bootloader are built
        # against macOS 11.x SDK). Therefore, Tcl shared library might end up unresolved (None); but that implicitly
        # indicates that the system framework is used.
        if tcl_shared_lib is None:
            return True

        # Check if the path corresponds to the system framework, i.e., [/System]/Library/Frameworks/Tcl.framework/Tcl
        return 'Library/Frameworks/Tcl.framework' in tcl_shared_lib

    @staticmethod
    def _warn_if_using_activetcl_or_teapot(tcl_root):
        """
        Check if Tcl installation is a Teapot-distributed version of ActiveTcl, and log a non-fatal warning that the
        resulting frozen application will (likely) fail to run on other systems.

        PyInstaller does *not* freeze all ActiveTcl dependencies -- including Teapot, which is typically ignorable.
        Since Teapot is *not* ignorable in this case, this function warns of impending failure.

        See Also
        -------
        https://github.com/pyinstaller/pyinstaller/issues/621
        """
        if tcl_root is None:
            return

        # Read the "init.tcl" script and look for mentions of "activetcl" and "teapot"
        init_tcl = os.path.join(tcl_root, 'init.tcl')
        if not os.path.isfile(init_tcl):
            return

        mentions_activetcl = False
        mentions_teapot = False

        # Tcl/Tk reads files using the system encoding (https://www.tcl.tk/doc/howto/i18n.html#system_encoding);
        # on macOS, this is UTF-8.
        with open(init_tcl, 'r', encoding='utf8') as fp:
            for line in fp.readlines():
                line = line.strip().lower()
                if line.startswith('#'):
                    continue
                if 'activetcl' in line:
                    mentions_activetcl = True
                if 'teapot' in line:
                    mentions_teapot = True
                if mentions_activetcl and mentions_teapot:
                    break

        if mentions_activetcl and mentions_teapot:
            logger.warning(
                "You appear to be using an ActiveTcl build of Tcl/Tk, which PyInstaller has\n"
                "difficulty freezing. To fix this, comment out all references to 'teapot' in\n"
                f"{init_tcl!r}\n"
                "See https://github.com/pyinstaller/pyinstaller/issues/621 for more information."
            )


tcltk_info = TclTkInfo()



#-----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import argparse
import os
import sys
import subprocess

if __name__ == '__main__':
    # Argument parser
    parser = argparse.ArgumentParser(description="Sub-process environment inheritance test")
    subparsers = parser.add_subparsers(dest="process_type", help="Process type.")

    parser_parent = subparsers.add_parser("parent", help="Parent process.")
    parser_parent.add_argument(
        "child_executable",
        type=str,
        help="Path to executable to spawn as a child process.",
    )
    parser_parent.add_argument(
        "--force-independent",
        action="store_true",
        help="Spawn the child process as an independent instance.",
    )

    parser_child = subparsers.add_parser("child", help="Child process.")
    parser_child.add_argument(
        "mode",
        type=str,
        choices={"same", "different"},
        help="Child mode - whether executable is the same as parent process or not.",
    )
    parser_child.add_argument(
        "parent_app_dir",
        type=str,
        help="Top-level application directory of the parent process.",
    )

    args = parser.parse_args()

    if args.process_type is None:
        # Process type not specified - exit early.
        print("Process type not specified.", file=sys.stderr)
    elif args.process_type == "parent":
        # Parent process.
        print("Running as parent process!", file=sys.stderr)
        print(f"Child executable: {args.child_executable}", file=sys.stderr)
        print(f"Force independent instance: {args.force_independent}", file=sys.stderr)

        env = os.environ.copy()

        if (
            args.child_executable == "sys.executable"
            or os.path.realpath(args.child_executable) == os.path.realpath(sys.executable)
        ):
            subprocess_args = [
                sys.executable,
                "child",
                "different" if args.force_independent else "same",
                sys._MEIPASS,
            ]

            if args.force_independent:
                env["PYINSTALLER_RESET_ENVIRONMENT"] = "1"
        else:
            subprocess_args = [
                args.child_executable,
                "child",
                "different",
                sys._MEIPASS,
            ]

        print(f"Running child process with arguments: {subprocess_args!r}", file=sys.stderr)
        subprocess.run(subprocess_args, check=True, env=env)
    else:
        # Child process.
        print("Running as child process!", file=sys.stderr)
        print(f"Mode: {args.mode}", file=sys.stderr)
        print(f"Our top-level directory: {sys._MEIPASS}", file=sys.stderr)
        print(f"Parent top-level directory: {args.parent_app_dir}", file=sys.stderr)

        if args.mode == "same":
            # Raise error if directories are different.
            if sys._MEIPASS != args.parent_app_dir:
                raise SystemExit(
                    f"Expected top-level directory ({sys._MEIPASS}) to be SAME as that of the parent process "
                    f"({args.parent_app_dir}), but it is not!"
                )
        else:
            # Raise error if directories are the same.
            if sys._MEIPASS == args.parent_app_dir:
                raise SystemExit(
                    f"Expected top-level directory ({sys._MEIPASS}) to be DIFFERENT from that of the parent process "
                    f"({args.parent_app_dir}), but it is not!"
                )



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Tests for splash screen.

NOTE: splash screen is not supported on macOS due to incompatible design.
"""

import pytest

from PyInstaller.compat import is_darwin, is_musl
from PyInstaller.utils.tests import importorskip

pytestmark = [
    importorskip('tkinter'),
    pytest.mark.skipif(is_darwin, reason="Splash screen is not supported on macOS."),
    pytest.mark.xfail(is_musl, reason="musl + tkinter is known to cause mysterious segfaults."),
]


# Test that splash screen is successfully started. This is an "interactive"
# test (see test_interactive.py), where we expect the program to run for
# specified amount of time, before we terminate it.
@pytest.mark.parametrize("build_mode", ['onedir', 'onefile'])
@pytest.mark.parametrize("with_tkinter", [False, True], ids=['notkinter', 'tkinter'])
def test_splash_screen_running(pyi_builder_spec, capfd, monkeypatch, build_mode, with_tkinter):
    if build_mode == 'onefile':
        monkeypatch.setenv('_TEST_SPLASH_BUILD_MODE', 'onefile')
    if with_tkinter:
        monkeypatch.setenv('_TEST_SPLASH_WITH_TKINTER', '1')

    pyi_builder_spec.test_spec(
        'spec_with_splash.spec',
        runtime=10,  # Interactive test - terminate test program after 10 seconds.
    )

    out, err = capfd.readouterr()
    assert 'SPLASH: splash screen started' in err, \
        f"Cannot find log entry indicating start of splash screen in:\n{err}"


# Check that splash-enabled program properly exits (e.g., there are no crashes in bootloader during shutdown).
# This is not coveved by the `test_splash_screen_running`, because that one terminates the program.
# There are three variants of this test:
# - splash screen is kept running until the end
# - user imports pyi_splash, which calls pyi_splash.close() via atexit()
# - user imports pyi_splash and calls pyi_splash.close()
def test_splash_screen_shutdown_auto(pyi_builder, script_dir):
    splash_image = script_dir.parent / 'data' / 'splash' / 'image.png'
    pyi_builder.test_source(
        """
        print("Done!")
        """,
        pyi_args=["--splash", str(splash_image)],
    )


def test_splash_screen_shutdown_atexit(pyi_builder, script_dir):
    splash_image = script_dir.parent / 'data' / 'splash' / 'image.png'
    pyi_builder.test_source(
        """
        # Importing pyi_splash registers pyi_splash.close() call via atexit().
        print("Importing pyi_splash...")
        import pyi_splash

        print("Done!")
        """,
        pyi_args=["--splash", str(splash_image)],
    )


def test_splash_screen_shutdown_manual(pyi_builder, script_dir):
    splash_image = script_dir.parent / 'data' / 'splash' / 'image.png'
    pyi_builder.test_source(
        """
        print("Importing pyi_splash...")
        import pyi_splash

        print("Closing splash screen from program...")
        pyi_splash.close()

        print("Done!")
        """,
        pyi_args=["--splash", str(splash_image)],
    )


# Check that splash screen is gracefully disabled in subprocesses spawned via sys.executable
def test_pyi_splash_in_subprocess(pyi_builder, script_dir):
    splash_image = script_dir.parent / 'data' / 'splash' / 'image.png'
    pyi_builder.test_script(
        "pyi_splash_in_subprocess.py",
        pyi_args=["--splash", str(splash_image)],
    )


# Check that splash screen is gracefully disabled via PYINSTALLER_SUPPRESS_SPLASH_SCREEN environment variable.
# The test procedure is adapted from pyi_splash_in_subprocess.py script of `test_pyi_splash_in_subprocess`.
def test_pyi_splash_suppress(pyi_builder, script_dir, monkeypatch):
    monkeypatch.setenv("PYINSTALLER_SUPPRESS_SPLASH_SCREEN", "1")
    splash_image = script_dir.parent / 'data' / 'splash' / 'image.png'
    pyi_builder.test_source(
        """
        import time

        # The pyi_splash module must be importable
        print("Importing pyi_splash...")
        import pyi_splash

        # Check module's guts to validate its state. Not for public use...

        # Module must be marked as initialized, regardless of whether splash screen is active or suppressed.
        assert pyi_splash._initialized, "Module is not marked as initialized!"

        # IPC port must be zero
        assert pyi_splash._ipc_port == 0, f"Unexpected IPC port value: {pyi_splash._ipc_port}"

        # Connection must be left closed
        assert pyi_splash._ipc_socket_closed, \
            "Unexpected splash screen socket state - expected it to be closed, but it is open!"

        # Test of public API

        # is_alive() should return false
        assert not pyi_splash.is_alive(), "Unexpected splash screen status!"

        # update_text() should be a no-op. Most importantly, it should not raise an error.
        pyi_splash.update_text("Updated text")
        time.sleep(1)

        # Same goes for close().
        pyi_splash.close()

        # After close, splash screen should be (still) inactive
        assert not pyi_splash.is_alive(), "Splash screen is not inactive!"
        """,
        pyi_args=["--splash", str(splash_image)],
    )



import unittest

import os, shutil, sys

from PyInstaller.lib.modulegraph import modulegraph

class ImpliesTestCase(unittest.TestCase):
    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, object, types, message=None):
            self.assertTrue(isinstance(object, types),
                    message or '%r is not an instance of %r'%(object, types))

    def testBasicImplies(self):
        root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-relimport')

        # First check that 'getopt' isn't accidentally in the graph:
        mg = modulegraph.ModuleGraph(path=[root]+sys.path)
        mg.add_script(os.path.join(root, 'script.py'))
        node = mg.find_node('mod')
        self.assertIsInstance(node, modulegraph.SourceModule)

        node = mg.find_node('getopt')
        self.assertEqual(node, None)

        # Now check that adding an implied dependency actually adds
        # 'getopt' to the graph:
        mg = modulegraph.ModuleGraph(path=[root]+sys.path, implies={
            'mod': ['getopt']})
        self.assertEqual(node, None)
        mg.add_script(os.path.join(root, 'script.py'))
        node = mg.find_node('mod')
        self.assertIsInstance(node, modulegraph.SourceModule)

        node = mg.find_node('getopt')
        self.assertIsInstance(node, modulegraph.SourceModule)

        # Check that the edges are correct:
        self.assertIn(mg.find_node('mod'), mg.get_edges(node)[1])
        self.assertIn(node, mg.get_edges(mg.find_node('mod'))[0])

    def testPackagedImplies(self):
        root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-relimport')

        # First check that 'getopt' isn't accidentally in the graph:
        mg = modulegraph.ModuleGraph(path=[root]+sys.path)
        mg.add_script(os.path.join(root, 'script.py'))
        node = mg.find_node('mod')
        self.assertIsInstance(node, modulegraph.SourceModule)

        node = mg.find_node('getopt')
        self.assertEqual(node, None)


        # Now check that adding an implied dependency actually adds
        # 'getopt' to the graph:
        mg = modulegraph.ModuleGraph(path=[root]+sys.path, implies={
            'pkg.relative': ['getopt']})
        node = mg.find_node('getopt')
        self.assertEqual(node, None)
        mg.add_script(os.path.join(root, 'script.py'))
        node = mg.find_node('pkg.relative')
        self.assertIsInstance(node, modulegraph.SourceModule)

        node = mg.find_node('getopt')
        self.assertIsInstance(node, modulegraph.SourceModule)

        # Check that the edges are correct:
        self.assertIn(mg.find_node('pkg.relative'), mg.get_edges(node)[1])
        self.assertIn(node, mg.get_edges(mg.find_node('pkg.relative'))[0])


if __name__ == '__main__':
    unittest.main()



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# These tests run a test script (scripts/pyi_importlib_resources.py) in unfrozen and frozen form, in combination with
# a custom test package (modules/pyi_pkg_resources_provider/package). In contrast to test_pkg_resources_provider.py,
# only source package form is used (zipped egg form is nor tested).
#
# Running the unfrozen test script allows us to verify the behavior of importlib.resources (or its importlib_resources
# back-port for python 3.8 and earlier) and thereby also validate the test script itself. Running the frozen test
# validates the behavior of the resource reader provided by PyInstaller's PyiFrozenLoader.
#
# For details on the structure of the test and the contents of the test package, see the top comment in the test script
# itself.

import os
import pathlib

import pytest

from PyInstaller.utils.tests import skipif
from PyInstaller.compat import is_darwin, is_py39, is_py310, exec_python_rc
from PyInstaller.utils.hooks import check_requirement

pytestmark = [
    skipif(
        not is_py39 and not check_requirement('importlib_resources'),
        reason="Python prior to 3.9 requires importlib_resources."
    )
]

# Directory with testing modules used in some tests.
_MODULES_DIR = pathlib.Path(__file__).parent / 'modules'


def __exec_python_script(script_filename, pathex):
    # Prepare the environment - default to 'os.environ'...
    env = os.environ.copy()
    # ... and prepend PYTHONPATH with pathex
    if 'PYTHONPATH' in env:
        pathex = os.pathsep.join([pathex, env['PYTHONPATH']])
    env['PYTHONPATH'] = pathex
    # Run the test script
    return exec_python_rc(script_filename, env=env)


def test_importlib_resources_source(script_dir):
    # Run the test script unfrozen - to validate it is working and to verify the behavior of importlib.resources
    # (or importlib_resources back-port).
    pathex = _MODULES_DIR / 'pyi_pkg_resources_provider' / 'package'
    test_script = script_dir / 'pyi_importlib_resources.py'
    ret = __exec_python_script(str(test_script), pathex=str(pathex))
    assert ret == 0, "Test script failed!"


def test_importlib_resources_frozen(pyi_builder, script_dir):
    # Run the test script as a frozen program
    pathex = _MODULES_DIR / 'pyi_pkg_resources_provider' / 'package'
    test_script = 'pyi_importlib_resources.py'
    hooks_dir = _MODULES_DIR / 'pyi_pkg_resources_provider' / 'hooks'
    pyi_args = [
        '--paths', str(pathex),
        '--hidden-import', 'pyi_pkgres_testpkg',
        '--additional-hooks-dir', str(hooks_dir),
    ]  # yapf: disable
    if is_darwin:
        pyi_args += ['--windowed']  # Also build and test .app bundle executable
    pyi_builder.test_script(
        test_script,
        pyi_args=pyi_args,
    )


# A separate test for verifying that `importlib.resources.files()` works with PEP-420 namespace packages. See #7921.
# The sub-directory containing the data files is also a PEP-410 namespace (sub)package. However, in the context of
# PyInstaller, there are actually two slightly different possibilities:
#  - if we collect only the data files, and the namespace package itself is not collected into PYZ, the situation is
#    the same as in unfrozen python - python's built-in import machinery takes care of the namespace package and the
#    associated resource reader.
#  - if the namespace package is collected into PYZ (in addition to resources being collected as data files), the
#    namespace package ends up being handled by PyInstaller's `PyiFrozenFinder`, which requires extra care to ensure
#    compatibility with `importlib` resource reader.
# The test covers both scenarios via `as_package` parameter.
#
# NOTE: the handling of resources in PEP-420 namespace packages as expected by this test was introduced in stdlib
# version of `importlib.resources` in python 3.10 (equivalent of importlib-resources >= 5.0).
@pytest.mark.parametrize('as_package', [True, False])
@skipif(
    not is_py310 and not check_requirement('importlib_resources >= 5.0'),
    reason="Requires python <= 3.10 or importlib_resources >= 5.0."
)
def test_importlib_resources_namespace_package_data_files(pyi_builder, as_package):
    pathex = _MODULES_DIR / 'pyi_namespace_package_with_data' / 'package'
    hooks_dir = _MODULES_DIR / 'pyi_namespace_package_with_data' / 'hooks'
    if as_package:
        hidden_imports = ['--hidden-import', 'pyi_test_nspkg', '--hidden-import', 'pyi_test_nspkg.data']
    else:
        hidden_imports = ['--hidden-import', 'pyi_test_nspkg']
    pyi_args = [
        '--paths', str(pathex),
        *hidden_imports,
        '--additional-hooks-dir', str(hooks_dir),
    ]  # yapf: disable
    if is_darwin:
        pyi_args += ['--windowed']  # Also build and test .app bundle executable
    pyi_builder.test_source(
        """
        import importlib
        try:
            import importlib_resources
        except ModuleNotFoundError:
            import importlib.resources as importlib_resources

        # Get the package's directory (= our data directory)
        data_dir = importlib_resources.files("pyi_test_nspkg.data")

        # Sanity check; verify the directory's base name
        assert data_dir.name == "data"

        # Check that data files exist
        assert (data_dir / "data_file1.txt").is_file()
        assert (data_dir / "data_file2.txt").is_file()
        assert (data_dir / "data_file3.txt").is_file()

        # Force cache invalidation and check again.
        importlib.invalidate_caches()

        data_dir = importlib_resources.files("pyi_test_nspkg.data")
        assert (data_dir / "data_file1.txt").is_file()
        """,
        pyi_args=pyi_args,
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: Apache-2.0
#-----------------------------------------------------------------------------
#
# The run-time hook provides a custom module iteration function for our PyiFrozenFinder, which allows
# `pkgutil.iter_modules()` to return entries for modules that are embedded in the PYZ archive. The non-embedded modules
# (binary extensions, modules collected as only source .py files, etc.) are enumerated using the `fallback_finder`
# provided by `PyiFrozenFinder` (which typically would be the python's `FileFinder`).
def _pyi_rthook():
    import pkgutil

    import pyimod02_importers  # PyInstaller's bootstrap module

    # This could, in fact, be implemented as `iter_modules()` method of the `PyiFrozenFinder`. However, we want to
    # avoid importing `pkgutil` in that bootstrap module (i.e., for the `pkgutil.iter_importer_modules()` call on the
    # fallback finder).
    def _iter_pyi_frozen_finder_modules(finder, prefix=''):
        # Fetch PYZ TOC tree from pyimod02_importers
        pyz_toc_tree = pyimod02_importers.get_pyz_toc_tree()

        # Finder has already pre-computed the package prefix implied by the search path. Use it to find the starting
        # node in the prefix tree.
        if finder._pyz_entry_prefix:
            pkg_name_parts = finder._pyz_entry_prefix.split('.')
        else:
            pkg_name_parts = []

        tree_node = pyz_toc_tree
        for pkg_name_part in pkg_name_parts:
            tree_node = tree_node.get(pkg_name_part)
            if not isinstance(tree_node, dict):
                # This check handles two cases:
                #  a) path does not exist (`tree_node` is `None`)
                #  b) path corresponds to a module instead of a package (`tree_node` is a leaf node (`str`)).
                tree_node = {}
                break

        # Dump the contents of the tree node.
        for entry_name, entry_data in tree_node.items():
            is_pkg = isinstance(entry_data, dict)
            yield prefix + entry_name, is_pkg

        # If our finder has a fall-back finder available, iterate its modules as well. By using the public
        # `fallback_finder` attribute, we force creation of the fallback finder as necessary.
        # NOTE: we do not care about potential duplicates here, because `pkgutil.iter_modules()` itself
        # keeps track of yielded names for purposes of de-duplication.
        if finder.fallback_finder is not None:
            yield from pkgutil.iter_importer_modules(finder.fallback_finder, prefix)

    pkgutil.iter_importer_modules.register(
        pyimod02_importers.PyiFrozenFinder,
        _iter_pyi_frozen_finder_modules,
    )


_pyi_rthook()
del _pyi_rthook



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import sys
import argparse
import pkgutil
import importlib

# Argument parser
parser = argparse.ArgumentParser(description="pkgutil iter_modules test")
parser.add_argument(
    'package',
    type=str,
    help="Package to test.",
)
parser.add_argument(
    '--prefix',
    type=str,
    default='',
    help="Optional prefix to pass to iter_modules.",
)
parser.add_argument(
    '--resolve-pkg-path',
    action='store_true',
    default=False,
    help="Resolve symbolic links in package path before passing it to pkgutil.iter_modules.",
)
parser.add_argument(
    '--output-file',
    default=None,
    type=str,
    help="Output file.",
)
args = parser.parse_args()

# Output file (optional)
if args.output_file:
    fp = open(args.output_file, 'w')
else:
    fp = sys.stdout

# Iterate over package's module
package = importlib.import_module(args.package)
pkg_path = package.__path__
if args.resolve_pkg_path:
    pkg_path = [os.path.realpath(path) for path in pkg_path]
for module in pkgutil.iter_modules(pkg_path, args.prefix):
    print("%s;%d" % (module.name, module.ispkg), file=fp)

# Cleanup
if args.output_file:
    fp.close()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Print all modules exporting the entry point 'pytest11'.
"""

import pkg_resources

plugins = sorted(i.module_name for i in pkg_resources.iter_entry_points("pytest11"))

print("\n".join(plugins))



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os.path
import sys

# Expand sys.path with PyInstaller source.
_ROOT_DIR = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..'))
sys.path.append(_ROOT_DIR)

# Bring all fixtures into this file.
from PyInstaller.utils.conftest import *  # noqa



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import array
import shutil

import pytest

from PyInstaller import HOMEPATH, PLATFORM
from PyInstaller.utils.misc import load_py_data_struct, save_py_data_struct


@pytest.mark.win32
def test_versioninfo(tmp_path):
    from PyInstaller.utils.win32.versioninfo import VSVersionInfo, \
        FixedFileInfo, StringFileInfo, StringTable, StringStruct, \
        VarFileInfo, VarStruct

    vsinfo = VSVersionInfo(
        ffi=FixedFileInfo(
            filevers=(1, 2, 3, 4),
            prodvers=(5, 6, 7, 8),
            mask=0x3f,
            flags=0x1,
            OS=0x40004,
            fileType=0x42,
            subtype=0x42,
            date=(0, 0)
        ),
        kids=[
            StringFileInfo([StringTable('040904b0', [StringStruct('FileDescription', 'versioninfo test')])]),
            VarFileInfo([VarStruct('Translation', [1033, 1200])])
        ]
    )

    file = str(tmp_path / 'versioninfo')
    save_py_data_struct(file, vsinfo)

    assert vsinfo == load_py_data_struct(file)


@pytest.mark.win32
def test_versioninfo_str(tmp_path):
    from PyInstaller.utils.win32.versioninfo import VSVersionInfo, \
        FixedFileInfo, StringFileInfo, StringTable, StringStruct, \
        VarFileInfo, VarStruct

    vsinfo = VSVersionInfo(
        ffi=FixedFileInfo(
            filevers=(1, 2, 3, 4),
            prodvers=(5, 6, 7, 8),
            mask=0x3f,
            flags=0x1,
            OS=0x40004,
            fileType=0x42,
            subtype=0x42,
            date=(0, 0)
        ),
        kids=[
            StringFileInfo([StringTable('040904b0', [StringStruct('FileDescription', 'versioninfo test')])]),
            VarFileInfo([VarStruct('Translation', [1033, 1200])])
        ]
    )

    # "Serialize" to string. This is what grab_version.py utility does to write VsVersionInfo to output text file.
    vs_info_str = str(vsinfo)

    # "Deserialize" via eval. This is what `versioninfo.load_version_info_from_text_file` does to read `VsVersionInfo`
    # from text file.
    vsinfo2 = eval(vs_info_str)

    assert vsinfo == vsinfo2


# Test that we properly serialize and deserialize VersionInfo that contains strings with quotes, such as the
# `FileDescription`  of `winrshost.exe` being `Host Process for WinRM's Remote Shell plugin`. See #7630.
@pytest.mark.win32
def test_versioninfo_str_quotes(tmp_path):
    from PyInstaller.utils.win32.versioninfo import VSVersionInfo, \
        FixedFileInfo, StringFileInfo, StringTable, StringStruct, \
        VarFileInfo, VarStruct

    FILE_DESCRIPTION = """versioninfo with quotes (' and ") test"""

    vsinfo = VSVersionInfo(
        ffi=FixedFileInfo(
            filevers=(1, 2, 3, 4),
            prodvers=(5, 6, 7, 8),
            mask=0x3f,
            flags=0x1,
            OS=0x40004,
            fileType=0x42,
            subtype=0x42,
            date=(0, 0)
        ),
        kids=[
            StringFileInfo([StringTable('040904b0', [StringStruct('FileDescription', FILE_DESCRIPTION)])]),
            VarFileInfo([VarStruct('Translation', [1033, 1200])])
        ]
    )

    # "Serialize" to string. This is what grab_version.py utility does to write VsVersionInfo to output text file.
    vs_info_str = str(vsinfo)

    # "Deserialize" via eval. This is what `versioninfo.load_version_info_from_text_file` does to read `VsVersionInfo`
    # from text file.
    vsinfo2 = eval(vs_info_str)

    assert vsinfo == vsinfo2


@pytest.mark.win32
def test_versioninfo_written_to_exe(tmp_path):
    from PyInstaller.utils.win32 import versioninfo
    from PyInstaller.utils.win32.versioninfo import VSVersionInfo, \
        FixedFileInfo, StringFileInfo, StringTable, StringStruct, \
        VarFileInfo, VarStruct

    # Test/expected values
    FILE_DESCRIPTION = 'versioninfo test'
    PRODUCT_NAME = 'Test Product'
    PRODUCT_VERSION = '2021.09.18.00'

    # Create a version info structure...
    vsinfo = VSVersionInfo(
        ffi=FixedFileInfo(
            filevers=(1, 2, 3, 4),
            prodvers=(5, 6, 7, 8),
            mask=0x3f,
            flags=0x1,
            OS=0x40004,
            fileType=0x42,
            subtype=0x42,
            date=(0, 0)
        ),
        kids=[
            StringFileInfo([
                StringTable(
                    '040904b0', [
                        StringStruct('FileDescription', FILE_DESCRIPTION),
                        StringStruct('ProductName', PRODUCT_NAME),
                        StringStruct('ProductVersion', PRODUCT_VERSION)
                    ]
                )
            ]),
            VarFileInfo([VarStruct('Translation', [1033, 1200])])
        ]
    )

    # Locate bootloader executable
    bootloader_file = os.path.join(HOMEPATH, 'PyInstaller', 'bootloader', PLATFORM, 'run.exe')

    # Create a local copy
    test_file = str(tmp_path / 'test_file.exe')
    shutil.copyfile(bootloader_file, test_file)

    # Embed version info
    versioninfo.write_version_info_to_executable(test_file, vsinfo)

    # Read back the values from the string table.
    def read_file_version_info(filename, *attributes):
        import ctypes
        winver = ctypes.WinDLL('version.dll')

        # Read the version information.
        info_len = winver.GetFileVersionInfoSizeW(filename, None)
        assert info_len, "No version information found!"
        info_buf = ctypes.create_string_buffer(info_len)
        rc = winver.GetFileVersionInfoW(filename, None, info_len, info_buf)
        assert rc, "Failed to read version information!"

        # Look for the codepages and take the first one.
        entry_ptr = ctypes.c_void_p()
        entry_len = ctypes.c_uint()
        winver.VerQueryValueW(info_buf, r'\VarFileInfo\Translation', ctypes.byref(entry_ptr), ctypes.byref(entry_len))
        assert entry_len.value, "No codepages found!"
        codepage = array.array('H', ctypes.string_at(entry_ptr.value, 4))
        codepage = "%04x%04x" % tuple(codepage)

        # Query attributes.
        attr_values = []
        for attr in attributes:
            entry_name = fr'\StringFileInfo\{codepage}\{attr}'
            rc = winver.VerQueryValueW(info_buf, entry_name, ctypes.byref(entry_ptr), ctypes.byref(entry_len))
            if not rc:
                entry_value = None
            else:
                entry_value = ctypes.wstring_at(entry_ptr.value, entry_len.value - 1)
            attr_values.append(entry_value)

        return attr_values

    values = read_file_version_info(test_file, 'FileDescription', 'ProductName', 'ProductVersion')
    assert values == [FILE_DESCRIPTION, PRODUCT_NAME, PRODUCT_VERSION]


CHINESE_LOREM_IPSUM = """
索感歴新策理通幻月続本初迷検価泉属速棄者。料流重者県作軽渡獲県行選見。意鑑民供変介画同表後院苦面例紙約度電心
時盤止野断希茶術同懲上権策全崎査応校匹国。真村辺名下周一場庭戸原図会正前対校米利。岩録開教純止表乗都偉別際政難然始京国京
校県水路中旅続球是校高約止性振下上派若福。案神整細番公流関閲元弊導説大実枝際確導
"""


def test_decode():
    """
    Test PyInstaller.utils.misc.decode().
    """
    import codecs
    from PyInstaller.utils.misc import decode

    # Test UTF8 with or without the BOM.
    assert decode(CHINESE_LOREM_IPSUM.encode()) == CHINESE_LOREM_IPSUM
    assert decode(codecs.BOM_UTF8 + CHINESE_LOREM_IPSUM.encode()) == CHINESE_LOREM_IPSUM

    # Test non-default UTF variants specified via a BOM.
    assert decode(codecs.BOM_UTF32_LE + codecs.utf_32_le_encode(CHINESE_LOREM_IPSUM)[0]) == CHINESE_LOREM_IPSUM
    assert decode(codecs.BOM_UTF16_BE + codecs.utf_16_be_encode(CHINESE_LOREM_IPSUM)[0]) == CHINESE_LOREM_IPSUM

    # Test using the encoding comment.
    with_cookie = "# encoding: gb18030\n" + CHINESE_LOREM_IPSUM
    assert decode(with_cookie.encode("GB18030")) == with_cookie



# -----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------
import subprocess


def test_collect_entry_point(pyi_builder_spec, script_dir, tmp_path):
    """
    Test PyInstaller.utils.hooks.collect_entry_point().

    On adding ``collect_entry_point('pytest11')`` to the spec file, the list of modules exporting the 'pytest11' entry
    point should be same after freezing.
    """
    from PyInstaller.compat import importlib_metadata
    plugins = sorted(i.module for i in importlib_metadata.entry_points(group="pytest11"))

    assert len(plugins), "The pytest11 entry point appears to have moved."

    pyi_builder_spec.test_spec('list_pytest11_entry_point.spec')
    exe = tmp_path / "dist" / "list_pytest11_entry_point" / "list_pytest11_entry_point"

    p = subprocess.run([str(exe)], stdout=subprocess.PIPE, check=True, encoding="utf-8")
    collected_plugins = p.stdout.strip("\n").split("\n")

    assert collected_plugins == plugins



"""
Test for import machinery
"""
import unittest
import sys
import textwrap
import subprocess
import os
from PyInstaller.lib.modulegraph import modulegraph

class TestModuleGraphImport (unittest.TestCase):
    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, value, types):
            if not isinstance(value, types):
                self.fail("%r is not an instance of %r"%(value, types))

    def setUp(self):
        self.root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-relimport2')
        self.mf = modulegraph.ModuleGraph(path=[ self.root ] + sys.path)


    def test_init_as_script(self):
        self.mf.add_script(os.path.join(self.root, 'pkg/__init__.py'))
        n = self.mf.find_node('mod1')
        self.assertIs(n, None)

        n = self.mf.find_node('.mod2.*')
        self.assertIsInstance(n, modulegraph.InvalidRelativeImport)

    def test_subpkg_bad_import(self):
        self.mf.import_hook('pkg.sub')

        n = self.mf.find_node('toplevel')
        self.assertIs(n, None)

        n = self.mf.find_node('pkg.mod1')
        self.assertIsInstance(n, modulegraph.SourceModule)

        n = self.mf.find_node('pkg.mod3')
        self.assertIsInstance(n, modulegraph.SourceModule)

if __name__ == "__main__":
    unittest.main()



# This hook should not be loaded at all.
# `mymodule_feature3` is imported by `mymodule_main.submodule_feature3` via import of `mymodule_feature3.submodule1`,
# but the hook for `mymodule_main` explicitly excludes `mymodule_feature3`. Therefore, if exclusion is applied correctly
# during processing of modules' imports (as opposed to a post-processing step), and if exclusion rules are applied in
# recursive way, this hook will not be loaded.
raise RuntimeError("This hook should not have been loaded!")



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import sys
import pprint
import _imp  # A built-in
import importlib

# NOTE: having `importlib` imported seems to ensure that `__file__` attribute is set on the `_frozen_importlib` module.
# Under python < 3.14, importing `pprint` seemed to have also imported `importlib`; but in 3.14.0a1, this is not the
# case. Therefore, we import `importlib` explicitly and use `importlib.import_module()` to load frozen modules (in
# earlier version of the test, `__import__()` was used). In contrast to `_frozen_importlib`, other frozen stdlib modules
# seem to have `__file__` set regardless of whether `importlib` is imported or not.

if sys.version_info < (3, 11):
    raise RuntimeError("Requires python >= 3.11!")

# Check that sys._stdlib_dir is set
if not sys._stdlib_dir:
    raise RuntimeError("sys._stdlib_dir is not set!")

# Frozen stdlib modules to test. Since these are frozen (by python itself), we do not need to specify them as hidden
# imports when this test script is frozen using PyInstaller.
frozen_stdlib_modules = sorted([name for name in _imp._frozen_module_names() if name in sys.stdlib_module_names])

output_data = [sys._stdlib_dir]  # First entry is sys._stdlib_dir
for module_name in frozen_stdlib_modules:
    print(f"Checking {module_name}...", file=sys.stderr)
    module = importlib.import_module(module_name)

    if not hasattr(module, '__file__'):
        raise RuntimeError(f"No __file__ attribute on {module_name}!")

    # Collect: module_name, __file__, filename and origname from __spec__.loaded_state
    loader_state = module.__spec__.loader_state
    entry = (
        module_name,
        module.__file__,
        loader_state.filename,
        loader_state.origname,
    )
    output_data.append(entry)

# Output: stdout or file
if len(sys.argv) > 1:
    with open(sys.argv[1], "w") as fp:
        pprint.pprint(output_data, stream=fp)
else:
    pprint.pprint(output_data)



import os

_forbidden_enabled = os.environ.get('_FORBIDDEN_MODULES_ENABLED', '1') == '1'

# Current level
from . import submod1  # noqa: F401, E402

try:
    from . import forbidden1  # noqa: F401
except ImportError:
    if _forbidden_enabled:
        raise

# Sub-level 1
from .subpkg1 import submod2  # noqa: F401, E402

try:
    from .subpkg1 import forbidden2  # noqa: F401
except ImportError:
    if _forbidden_enabled:
        raise

# Sub-level 2
from .subpkg1.subpkg2 import submod3  # noqa: F401, E402

try:
    from .subpkg1.subpkg2 import forbidden3  # noqa: F401
except ImportError:
    if _forbidden_enabled:
        raise



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib import Utils, Logs
from waflib.Tools import fc

fc_compiler = {
    'win32': ['gfortran', 'ifort'],
    'darwin': ['gfortran', 'g95', 'ifort'],
    'linux': ['gfortran', 'g95', 'ifort'],
    'java': ['gfortran', 'g95', 'ifort'],
    'default': ['gfortran'],
    'aix': ['gfortran']
}


def default_compilers():
    build_platform = Utils.unversioned_sys_platform()
    possible_compiler_list = fc_compiler.get(build_platform, fc_compiler['default'])
    return ' '.join(possible_compiler_list)


def configure(conf):
    try:
        test_for_compiler = conf.options.check_fortran_compiler or default_compilers()
    except AttributeError:
        conf.fatal("Add options(opt): opt.load('compiler_fc')")
    for compiler in re.split('[ ,]+', test_for_compiler):
        conf.env.stash()
        conf.start_msg('Checking for %r (Fortran compiler)' % compiler)
        try:
            conf.load(compiler)
        except conf.errors.ConfigurationError as e:
            conf.env.revert()
            conf.end_msg(False)
            Logs.debug('compiler_fortran: %r', e)
        else:
            if conf.env.FC:
                conf.end_msg(conf.env.get_flat('FC'))
                conf.env.COMPILER_FORTRAN = compiler
                conf.env.commit()
                break
            conf.env.revert()
            conf.end_msg(False)
    else:
        conf.fatal('could not configure a Fortran compiler!')


def options(opt):
    test_for_compiler = default_compilers()
    opt.load_special_tools('fc_*.py')
    fortran_compiler_opts = opt.add_option_group('Configuration options')
    fortran_compiler_opts.add_option(
        '--check-fortran-compiler',
        default=None,
        help='list of Fortran compiler to try [%s]' % test_for_compiler,
        dest="check_fortran_compiler"
    )
    for x in test_for_compiler.split():
        opt.load('%s' % x)



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Verify packaging of Sphinx, which relies on jinja2 and on docutils. Sphinx and docutils rely on data files in their
# module directories, which their respective hook scripts must find and copy.

import os
import sys

import sphinx.cmd.build

# The path to source data directory is passed via first command-line argument.
if len(sys.argv) != 2:
    print(f"Use: {sys.argv[0]} <data-dir>")
    raise SystemExit(1)
sphinx_path = sys.argv[1]

# Invoke Sphinx. See http://sphinx-doc.org/invocation.html#invocation-of-sphinx-build for more details
# on the used options.
ret = sphinx.cmd.build.main([
    # Rebuild all files.
    '-a', '-E',
    # Produce html output.
    '-b', 'html',
    # Specify an output directory for data files.
    '-d', os.path.join(sphinx_path, '_build', 'doctrees'),
    # Specify the location of the source (index.rst).
    sphinx_path,
    # Build directory for the resulting HTML files.
    os.path.join(sphinx_path, '_build', 'html')
])  # yapf: disable
raise SystemExit(ret)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re
from waflib import Utils, Task, Errors, Logs, Node
from waflib.TaskGen import feature, before_method

re_bibunit = re.compile(r'\\(?P<type>putbib)\[(?P<file>[^\[\]]*)\]', re.M)


def bibunitscan(self):
    node = self.inputs[0]
    nodes = []
    if not node:
        return nodes
    code = node.read()
    for match in re_bibunit.finditer(code):
        path = match.group('file')
        if path:
            found = None
            for k in ('', '.bib'):
                Logs.debug('tex: trying %s%s', path, k)
                fi = node.parent.find_resource(path + k)
                if fi:
                    found = True
                    nodes.append(fi)
            if not found:
                Logs.debug('tex: could not find %s', path)
    Logs.debug('tex: found the following bibunit files: %s', nodes)
    return nodes


exts_deps_tex = ['', '.ltx', '.tex', '.bib', '.pdf', '.png', '.eps', '.ps', '.sty']
exts_tex = ['.ltx', '.tex']
re_tex = re.compile(
    r'\\(?P<type>usepackage|RequirePackage|include|bibliography([^\[\]{}]*)|putbib|includegraphics|input|import|bringin|lstinputlisting)(\[[^\[\]]*\])?{(?P<file>[^{}]*)}',
    re.M
)
g_bibtex_re = re.compile('bibdata', re.M)
g_glossaries_re = re.compile('\\@newglossary', re.M)


class tex(Task.Task):
    bibtex_fun, _ = Task.compile_fun('${BIBTEX} ${BIBTEXFLAGS} ${SRCFILE}', shell=False)
    bibtex_fun.__doc__ = """
	Execute the program **bibtex**
	"""
    makeindex_fun, _ = Task.compile_fun('${MAKEINDEX} ${MAKEINDEXFLAGS} ${SRCFILE}', shell=False)
    makeindex_fun.__doc__ = """
	Execute the program **makeindex**
	"""
    makeglossaries_fun, _ = Task.compile_fun('${MAKEGLOSSARIES} ${SRCFILE}', shell=False)
    makeglossaries_fun.__doc__ = """
	Execute the program **makeglossaries**
	"""

    def exec_command(self, cmd, **kw):
        if self.env.PROMPT_LATEX:
            kw['stdout'] = kw['stderr'] = None
        return super(tex, self).exec_command(cmd, **kw)

    def scan_aux(self, node):
        nodes = [node]
        re_aux = re.compile(r'\\@input{(?P<file>[^{}]*)}', re.M)

        def parse_node(node):
            code = node.read()
            for match in re_aux.finditer(code):
                path = match.group('file')
                found = node.parent.find_or_declare(path)
                if found and found not in nodes:
                    Logs.debug('tex: found aux node %r', found)
                    nodes.append(found)
                    parse_node(found)

        parse_node(node)
        return nodes

    def scan(self):
        node = self.inputs[0]
        nodes = []
        names = []
        seen = []
        if not node:
            return (nodes, names)

        def parse_node(node):
            if node in seen:
                return
            seen.append(node)
            code = node.read()
            for match in re_tex.finditer(code):
                multibib = match.group('type')
                if multibib and multibib.startswith('bibliography'):
                    multibib = multibib[len('bibliography'):]
                    if multibib.startswith('style'):
                        continue
                else:
                    multibib = None
                for path in match.group('file').split(','):
                    if path:
                        add_name = True
                        found = None
                        for k in exts_deps_tex:
                            for up in self.texinputs_nodes:
                                Logs.debug('tex: trying %s%s', path, k)
                                found = up.find_resource(path + k)
                                if found:
                                    break
                            for tsk in self.generator.tasks:
                                if not found or found in tsk.outputs:
                                    break
                            else:
                                nodes.append(found)
                                add_name = False
                                for ext in exts_tex:
                                    if found.name.endswith(ext):
                                        parse_node(found)
                                        break
                            if found and multibib and found.name.endswith('.bib'):
                                try:
                                    self.multibibs.append(found)
                                except AttributeError:
                                    self.multibibs = [found]
                        if add_name:
                            names.append(path)

        parse_node(node)
        for x in nodes:
            x.parent.get_bld().mkdir()
        Logs.debug("tex: found the following : %s and names %s", nodes, names)
        return (nodes, names)

    def check_status(self, msg, retcode):
        if retcode != 0:
            raise Errors.WafError('%r command exit status %r' % (msg, retcode))

    def info(self, *k, **kw):
        try:
            info = self.generator.bld.conf.logger.info
        except AttributeError:
            info = Logs.info
        info(*k, **kw)

    def bibfile(self):
        for aux_node in self.aux_nodes:
            try:
                ct = aux_node.read()
            except EnvironmentError:
                Logs.error('Error reading %s: %r', aux_node.abspath())
                continue
            if g_bibtex_re.findall(ct):
                self.info('calling bibtex')
                self.env.env = {}
                self.env.env.update(os.environ)
                self.env.env.update({'BIBINPUTS': self.texinputs(), 'BSTINPUTS': self.texinputs()})
                self.env.SRCFILE = aux_node.name[:-4]
                self.check_status('error when calling bibtex', self.bibtex_fun())
        for node in getattr(self, 'multibibs', []):
            self.env.env = {}
            self.env.env.update(os.environ)
            self.env.env.update({'BIBINPUTS': self.texinputs(), 'BSTINPUTS': self.texinputs()})
            self.env.SRCFILE = node.name[:-4]
            self.check_status('error when calling bibtex', self.bibtex_fun())

    def bibunits(self):
        try:
            bibunits = bibunitscan(self)
        except OSError:
            Logs.error('error bibunitscan')
        else:
            if bibunits:
                fn = ['bu' + str(i) for i in range(1, len(bibunits) + 1)]
                if fn:
                    self.info('calling bibtex on bibunits')
                for f in fn:
                    self.env.env = {'BIBINPUTS': self.texinputs(), 'BSTINPUTS': self.texinputs()}
                    self.env.SRCFILE = f
                    self.check_status('error when calling bibtex', self.bibtex_fun())

    def makeindex(self):
        self.idx_node = self.inputs[0].change_ext('.idx')
        try:
            idx_path = self.idx_node.abspath()
            os.stat(idx_path)
        except OSError:
            self.info('index file %s absent, not calling makeindex', idx_path)
        else:
            self.info('calling makeindex')
            self.env.SRCFILE = self.idx_node.name
            self.env.env = {}
            self.check_status('error when calling makeindex %s' % idx_path, self.makeindex_fun())

    def bibtopic(self):
        p = self.inputs[0].parent.get_bld()
        if os.path.exists(os.path.join(p.abspath(), 'btaux.aux')):
            self.aux_nodes += p.ant_glob('*[0-9].aux')

    def makeglossaries(self):
        src_file = self.inputs[0].abspath()
        base_file = os.path.basename(src_file)
        base, _ = os.path.splitext(base_file)
        for aux_node in self.aux_nodes:
            try:
                ct = aux_node.read()
            except EnvironmentError:
                Logs.error('Error reading %s: %r', aux_node.abspath())
                continue
            if g_glossaries_re.findall(ct):
                if not self.env.MAKEGLOSSARIES:
                    raise Errors.WafError("The program 'makeglossaries' is missing!")
                Logs.warn('calling makeglossaries')
                self.env.SRCFILE = base
                self.check_status('error when calling makeglossaries %s' % base, self.makeglossaries_fun())
                return

    def texinputs(self):
        return os.pathsep.join([k.abspath() for k in self.texinputs_nodes]) + os.pathsep

    def run(self):
        env = self.env
        if not env.PROMPT_LATEX:
            env.append_value('LATEXFLAGS', '-interaction=batchmode')
            env.append_value('PDFLATEXFLAGS', '-interaction=batchmode')
            env.append_value('XELATEXFLAGS', '-interaction=batchmode')
        self.cwd = self.inputs[0].parent.get_bld()
        self.info('first pass on %s', self.__class__.__name__)
        cur_hash = self.hash_aux_nodes()
        self.call_latex()
        self.hash_aux_nodes()
        self.bibtopic()
        self.bibfile()
        self.bibunits()
        self.makeindex()
        self.makeglossaries()
        for i in range(10):
            prev_hash = cur_hash
            cur_hash = self.hash_aux_nodes()
            if not cur_hash:
                Logs.error('No aux.h to process')
            if cur_hash and cur_hash == prev_hash:
                break
            self.info('calling %s', self.__class__.__name__)
            self.call_latex()

    def hash_aux_nodes(self):
        try:
            self.aux_nodes
        except AttributeError:
            try:
                self.aux_nodes = self.scan_aux(self.inputs[0].change_ext('.aux'))
            except IOError:
                return None
        return Utils.h_list([Utils.h_file(x.abspath()) for x in self.aux_nodes])

    def call_latex(self):
        self.env.env = {}
        self.env.env.update(os.environ)
        self.env.env.update({'TEXINPUTS': self.texinputs()})
        self.env.SRCFILE = self.inputs[0].abspath()
        self.check_status('error when calling latex', self.texfun())


class latex(tex):
    texfun, vars = Task.compile_fun('${LATEX} ${LATEXFLAGS} ${SRCFILE}', shell=False)


class pdflatex(tex):
    texfun, vars = Task.compile_fun('${PDFLATEX} ${PDFLATEXFLAGS} ${SRCFILE}', shell=False)


class xelatex(tex):
    texfun, vars = Task.compile_fun('${XELATEX} ${XELATEXFLAGS} ${SRCFILE}', shell=False)


class dvips(Task.Task):
    run_str = '${DVIPS} ${DVIPSFLAGS} ${SRC} -o ${TGT}'
    color = 'BLUE'
    after = ['latex', 'pdflatex', 'xelatex']


class dvipdf(Task.Task):
    run_str = '${DVIPDF} ${DVIPDFFLAGS} ${SRC} ${TGT}'
    color = 'BLUE'
    after = ['latex', 'pdflatex', 'xelatex']


class pdf2ps(Task.Task):
    run_str = '${PDF2PS} ${PDF2PSFLAGS} ${SRC} ${TGT}'
    color = 'BLUE'
    after = ['latex', 'pdflatex', 'xelatex']


@feature('tex')
@before_method('process_source')
def apply_tex(self):
    if not getattr(self, 'type', None) in ('latex', 'pdflatex', 'xelatex'):
        self.type = 'pdflatex'
    outs = Utils.to_list(getattr(self, 'outs', []))
    try:
        self.generator.bld.conf
    except AttributeError:
        default_prompt = False
    else:
        default_prompt = True
    self.env.PROMPT_LATEX = getattr(self, 'prompt', default_prompt)
    deps_lst = []
    if getattr(self, 'deps', None):
        deps = self.to_list(self.deps)
        for dep in deps:
            if isinstance(dep, str):
                n = self.path.find_resource(dep)
                if not n:
                    self.bld.fatal('Could not find %r for %r' % (dep, self))
                if not n in deps_lst:
                    deps_lst.append(n)
            elif isinstance(dep, Node.Node):
                deps_lst.append(dep)
    for node in self.to_nodes(self.source):
        if self.type == 'latex':
            task = self.create_task('latex', node, node.change_ext('.dvi'))
        elif self.type == 'pdflatex':
            task = self.create_task('pdflatex', node, node.change_ext('.pdf'))
        elif self.type == 'xelatex':
            task = self.create_task('xelatex', node, node.change_ext('.pdf'))
        task.env = self.env
        if deps_lst:
            for n in deps_lst:
                if not n in task.dep_nodes:
                    task.dep_nodes.append(n)
        if hasattr(self, 'texinputs_nodes'):
            task.texinputs_nodes = self.texinputs_nodes
        else:
            task.texinputs_nodes = [node.parent, node.parent.get_bld(), self.path, self.path.get_bld()]
            lst = os.environ.get('TEXINPUTS', '')
            if self.env.TEXINPUTS:
                lst += os.pathsep + self.env.TEXINPUTS
            if lst:
                lst = lst.split(os.pathsep)
            for x in lst:
                if x:
                    if os.path.isabs(x):
                        p = self.bld.root.find_node(x)
                        if p:
                            task.texinputs_nodes.append(p)
                        else:
                            Logs.error('Invalid TEXINPUTS folder %s', x)
                    else:
                        Logs.error('Cannot resolve relative paths in TEXINPUTS %s', x)
        if self.type == 'latex':
            if 'ps' in outs:
                tsk = self.create_task('dvips', task.outputs, node.change_ext('.ps'))
                tsk.env.env = dict(os.environ)
            if 'pdf' in outs:
                tsk = self.create_task('dvipdf', task.outputs, node.change_ext('.pdf'))
                tsk.env.env = dict(os.environ)
        elif self.type == 'pdflatex':
            if 'ps' in outs:
                self.create_task('pdf2ps', task.outputs, node.change_ext('.ps'))
    self.source = []


def configure(self):
    v = self.env
    for p in 'tex latex pdflatex xelatex bibtex dvips dvipdf ps2pdf makeindex pdf2ps makeglossaries'.split():
        try:
            self.find_program(p, var=p.upper())
        except self.errors.ConfigurationError:
            pass
    v.DVIPSFLAGS = '-Ppdf'




import toplevel_existing
import toplevel_nonexisting

class MyClass:
    import toplevel_class_existing
    import toplevel_class_nonexisting

if a == b:
    import toplevel_conditional_existing
    import toplevel_conditional_nonexisting

    try:
        import toplevel_conditional_import_existing
        import toplevel_conditional_import_nonexisting
    except:
        import toplevel_conditional_import2_existing
        import toplevel_conditional_import2_nonexisting

try:
    import toplevel_import_existing
    import toplevel_import_nonexisting
except:
    import toplevel_import2_existing
    import toplevel_import2_nonexisting

def function():
    import function_existing
    import function_nonexisting

    class MyClass:
        import function_class_existing
        import function_class_nonexisting

    if a == b:
        import function_conditional_existing
        import function_conditional_nonexisting

        try:
            import function_conditional_import_existing
            import function_conditional_import_nonexisting
        except:
            import function_conditional_import2_existing
            import function_conditional_import2_nonexisting

    try:
        import function_import_existing
        import function_import_nonexisting
    except:
        import function_import2_existing
        import function_import2_nonexisting



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re
from waflib import Task, TaskGen
from waflib.Tools import ccroot


def decide_ext(self, node):
    if 'cxx' in self.features:
        return ['.lex.cc']
    return ['.lex.c']


def flexfun(tsk):
    env = tsk.env
    bld = tsk.generator.bld
    wd = bld.variant_dir

    def to_list(xx):
        if isinstance(xx, str):
            return [xx]
        return xx

    tsk.last_cmd = lst = []
    lst.extend(to_list(env.FLEX))
    lst.extend(to_list(env.FLEXFLAGS))
    inputs = [a.path_from(tsk.get_cwd()) for a in tsk.inputs]
    if env.FLEX_MSYS:
        inputs = [x.replace(os.sep, '/') for x in inputs]
    lst.extend(inputs)
    lst = [x for x in lst if x]
    txt = bld.cmd_and_log(lst, cwd=wd, env=env.env or None, quiet=0)
    tsk.outputs[0].write(txt.replace('\r\n', '\n').replace('\r', '\n'))


TaskGen.declare_chain(
    name='flex',
    rule=flexfun,
    ext_in='.l',
    decider=decide_ext,
)
Task.classes['flex'].vars = ['FLEXFLAGS', 'FLEX']
ccroot.USELIB_VARS['c'].add('FLEXFLAGS')
ccroot.USELIB_VARS['cxx'].add('FLEXFLAGS')


def configure(conf):
    conf.find_program('flex', var='FLEX')
    conf.env.FLEXFLAGS = ['-t']
    if re.search(r"\\msys\\[0-9.]+\\bin\\flex.exe$", conf.env.FLEX[0]):
        conf.env.FLEX_MSYS = True



import http.client

def div(a, b):
    try:
        return a/b

    except ZeroDivisionError as exc:
        print("Zero division!")
        return None

class MyClass (object, metaclass=type):
    pass



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Utils, Task, Errors
from waflib.TaskGen import taskgen_method, feature, extension
from waflib.Tools import d_scan, d_config
from waflib.Tools.ccroot import link_task, stlink_task


class d(Task.Task):
    color = 'GREEN'
    run_str = '${D} ${DFLAGS} ${DINC_ST:INCPATHS} ${D_SRC_F:SRC} ${D_TGT_F:TGT}'
    scan = d_scan.scan


class d_with_header(d):
    run_str = '${D} ${DFLAGS} ${DINC_ST:INCPATHS} ${D_HDR_F:tgt.outputs[1].bldpath()} ${D_SRC_F:SRC} ${D_TGT_F:tgt.outputs[0].bldpath()}'


class d_header(Task.Task):
    color = 'BLUE'
    run_str = '${D} ${D_HEADER} ${SRC}'


class dprogram(link_task):
    run_str = '${D_LINKER} ${LINKFLAGS} ${DLNK_SRC_F}${SRC} ${DLNK_TGT_F:TGT} ${RPATH_ST:RPATH} ${DSTLIB_MARKER} ${DSTLIBPATH_ST:STLIBPATH} ${DSTLIB_ST:STLIB} ${DSHLIB_MARKER} ${DLIBPATH_ST:LIBPATH} ${DSHLIB_ST:LIB}'
    inst_to = '${BINDIR}'


class dshlib(dprogram):
    inst_to = '${LIBDIR}'


class dstlib(stlink_task):
    pass


@extension('.d', '.di', '.D')
def d_hook(self, node):
    ext = Utils.destos_to_binfmt(self.env.DEST_OS) == 'pe' and 'obj' or 'o'
    out = '%s.%d.%s' % (node.name, self.idx, ext)

    def create_compiled_task(self, name, node):
        task = self.create_task(name, node, node.parent.find_or_declare(out))
        try:
            self.compiled_tasks.append(task)
        except AttributeError:
            self.compiled_tasks = [task]
        return task

    if getattr(self, 'generate_headers', None):
        tsk = create_compiled_task(self, 'd_with_header', node)
        tsk.outputs.append(node.change_ext(self.env.DHEADER_ext))
    else:
        tsk = create_compiled_task(self, 'd', node)
    return tsk


@taskgen_method
def generate_header(self, filename):
    try:
        self.header_lst.append([filename, self.install_path])
    except AttributeError:
        self.header_lst = [[filename, self.install_path]]


@feature('d')
def process_header(self):
    for i in getattr(self, 'header_lst', []):
        node = self.path.find_resource(i[0])
        if not node:
            raise Errors.WafError('file %r not found on d obj' % i[0])
        self.create_task('d_header', node, node.change_ext('.di'))



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Functional tests for the Django content management system (CMS).
"""

from PyInstaller.utils.tests import importorskip


@importorskip('django')
def test_django(pyi_builder, monkeypatch, data_dir):
    # Extend sys.path so PyInstaller could find modules from 'tmpdir/django/'.
    monkeypatch.syspath_prepend(str(data_dir))
    # Django uses manage.py as the main script.
    script = data_dir / 'manage.py'
    # Create the exe, run django command 'check' to do basic sanity checking of the executable.
    pyi_builder.test_script(script, app_name='django_site', app_args=['check'])



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import sys
import signal
import subprocess

import pytest

from PyInstaller.utils.tests import onefile_only


@pytest.mark.darwin
@pytest.mark.linux
@pytest.mark.parametrize('forward_signals', [True, False], ids=['forward', 'ignore'])
@onefile_only
def test_onefile_signal_handling(pyi_builder, forward_signals):
    # Build the test program. The `pyi_builder.test_soruce` also runs the built program, but since no arguments are
    # passed via command-line, this program run is a no-op.
    pyi_builder.test_source(
        """
        import os
        import sys
        import signal
        import time

        # Quietly exit if no signal number is given on command-line. This accommodates the fact that
        # `pyi_builder.test_source()` always runs the program after building it.
        if len(sys.argv) < 3:
            print(f"Usage: {sys.argv[0]} <signal-number> <timeout>", file=sys.stderr)
            sys.exit(0)

        # Signal number is passed as the first command-line argument
        signal_number = int(sys.argv[1])
        signal_name = signal.Signals(signal_number).name  # This implicitly validates signal number

        # Timeout is passed as the second command-line argument
        timeout = float(sys.argv[2])

        # Return code: received signal number, or zero if signal was not received.
        return_code = 0

        # Install signal handler
        def signal_handler(signum, *args):
            global return_code
            return_code = signum  # Set program's return code to signal number

        print(f"Installing signal handler for signal={signal_number} ({signal_name})", file=sys.stderr)
        signal.signal(signal_number, signal_handler)

        # Send signal to parent process of the onefile frozen application.
        parent_pid = os.getppid()
        print(
            f"Sending signal={signal_number} ({signal_name}) to parent process with PID={parent_pid}",
            file=sys.stderr,
        )
        os.kill(parent_pid, signal_number)

        # Wait for signal to be delivered or the specified timeout interval to pass.
        start_time = time.time()
        while True:
            elapsed_time = time.time() - start_time
            if elapsed_time >= timeout:
                print(f"Signal not received within {timeout} seconds!", file=sys.stderr)
                break
            if return_code != 0:
                print(f"Signal received! Elapsed time: {elapsed_time:.2f} seconds.", file=sys.stderr)
                break
            time.sleep(0.1)  # 100 ms steps

        sys.exit(return_code)
        """,
        pyi_args=[] if forward_signals else ['--bootloader-ignore-signals'],
    )

    exes = pyi_builder._find_executables('test_source')
    assert len(exes) == 1
    program_exe = exes[0]

    # Use the built executable with all applicable signals.
    failures = []
    for signal_entry in signal.Signals:
        signal_name = signal_entry.name
        signal_number = signal_entry.value

        # Exemptions
        reason = None
        if signal_name in {'SIGKILL', 'SIGSTOP'}:
            reason = f"{signal_name} cannot be caught."
        elif signal_name in {'SIGCHLD', 'SIGCLD'}:
            reason = f"{signal_name} is not handled by bootloader: required for wait() on child process."
        elif signal_name == 'SIGTSTP':
            reason = f"{signal_name} is not handled by bootloader: required for Ctrl-Z."

        if reason is not None:
            print(f"=== skipping test with {signal_name}: {reason} ===", file=sys.stderr)
            continue

        # Expected return code: signal number in signal-forwarding mode, zero otherwise.
        expected_code = signal_number if forward_signals else 0

        # Timeout interval when waiting for signal to be delivered (or not): in signal-forwarding mode, wait up to 5
        # seconds, to be on the safe side and avoid sporadic test failures. The initial 1-second interval is sometimes
        # too short when running tests under CPU contention scenario (number of pytest runners matching or exceeding
        # number of CPU cores). Since the test program exits as soon as signal is received, we expect to never hit this
        # 5-seconds limit. In contrast, when in signal-ignoring mode, we need to wait for the whole interval to see that
        # the signal is not delivered. In this case, we stick with original 1-second timeout to avoid unnecessarily
        # slowing down the test; as mentioned earlier, we expect signals to be delivered within 1 second for most of the
        # time, so if ignore mode was broken, at least some tests would fail.
        timeout = 5 if forward_signals else 1  # seconds

        # Run
        print(f"=== running test with {signal_name} ===", file=sys.stderr)
        status = subprocess.run([program_exe, str(signal_number), str(timeout)])
        ret_code = status.returncode

        print(f"=== program returned code {ret_code} (expected {expected_code}) ===", file=sys.stderr)
        if ret_code != expected_code:
            failures.append((signal_name, f"Unexpected exit code: {ret_code} (expected={expected_code})!"))

    assert not failures, "Not all signals were handled as expected!"



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------


def pre_safe_import_module(api):
    # PyGObject modules loaded through the gi repository are marked as MissingModules by modulegraph, so we convert them
    # to RuntimeModules in order for their hooks to be loaded and executed.
    api.add_runtime_module(api.module_name)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.Tools import ccroot, ar
from waflib.Configure import conf


@conf
def find_xlc(conf):
    cc = conf.find_program(['xlc_r', 'xlc'], var='CC')
    conf.get_xlc_version(cc)
    conf.env.CC_NAME = 'xlc'


@conf
def xlc_common_flags(conf):
    v = conf.env
    v.CC_SRC_F = []
    v.CC_TGT_F = ['-c', '-o']
    if not v.LINK_CC:
        v.LINK_CC = v.CC
    v.CCLNK_SRC_F = []
    v.CCLNK_TGT_F = ['-o']
    v.CPPPATH_ST = '-I%s'
    v.DEFINES_ST = '-D%s'
    v.LIB_ST = '-l%s'
    v.LIBPATH_ST = '-L%s'
    v.STLIB_ST = '-l%s'
    v.STLIBPATH_ST = '-L%s'
    v.RPATH_ST = '-Wl,-rpath,%s'
    v.SONAME_ST = []
    v.SHLIB_MARKER = []
    v.STLIB_MARKER = []
    v.LINKFLAGS_cprogram = ['-Wl,-brtl']
    v.cprogram_PATTERN = '%s'
    v.CFLAGS_cshlib = ['-fPIC']
    v.LINKFLAGS_cshlib = ['-G', '-Wl,-brtl,-bexpfull']
    v.cshlib_PATTERN = 'lib%s.so'
    v.LINKFLAGS_cstlib = []
    v.cstlib_PATTERN = 'lib%s.a'


def configure(conf):
    conf.find_xlc()
    conf.find_ar()
    conf.xlc_common_flags()
    conf.cc_load_tools()
    conf.cc_add_flags()
    conf.link_add_flags()



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------


# A basic lazy loader test with a stdlib module
def test_importlib_lazy_loader(pyi_builder):
    pyi_builder.test_script(
        'pyi_lazy_import.py',
        app_args=['json'],
        pyi_args=['--hiddenimport', 'json'],
    )


# NOTE: the tests with aliased module used to be based on `pkg_resources._vendor.jaraco.text` and
# `pkg_resources.extern.jaraco.text`. However, `pkg_resources` shipped with `setuptools` >= 71  does not vendor its
# dependencies anymore, so those modules are gone. Therefore, the tests are now using `pyi_testmod_metapath1` from
# `test_import_metapath1`, which implements a copy of `VendorImporter` used by earlier versions of `setuptools`.


# Lazy loader test with aliased module - using original name
def test_importlib_lazy_loader_alias1(pyi_builder, script_dir):
    pyi_builder.test_script(
        'pyi_lazy_import.py',
        app_args=['pyi_testmod_metapath1._vendor.ccc.ddd'],
        pyi_args=[
            '--hiddenimport',
            'pyi_testmod_metapath1',
            '--additional-hooks-dir',
            str(script_dir / 'pyi_hooks'),
        ],
    )


# Lazy loader test with aliased module - using alias
def test_importlib_lazy_loader_alias2(pyi_builder, script_dir):
    pyi_builder.test_script(
        'pyi_lazy_import.py',
        app_args=['pyi_testmod_metapath1.extern.ccc.ddd'],
        pyi_args=[
            '--hiddenimport',
            'pyi_testmod_metapath1',
            '--additional-hooks-dir',
            str(script_dir / 'pyi_hooks'),
        ],
    )



#!/usr/bin/env python3
#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Verify that new news entries have valid filenames. Usage:

.. code-block:: bash

    ./scripts/verify-news-fragments.py

"""

import re
import sys
from pathlib import Path

CHANGELOG_GUIDE = (
    "https://github.com/pyinstaller/pyinstaller/"
    "blob/develop/doc/development/changelog-entries.rst#changelog-entries"
)

CHANGE_TYPES = {
    'bootloader', 'breaking', 'bugfix', 'build', 'core', 'doc', 'feature', 'hooks', 'moduleloader', 'process', 'tests',
    'deprecation'
}

NEWS_PATTERN = re.compile(r"(\d+)\.(\w+)\.(?:(\d+)\.)?rst")

NEWS_DIR = Path(__file__).absolute().parent.parent / "news"


def validate_name(name):
    """
    Check a filename/filepath matches the required format.

    :param name: Name of news fragment file.
    :type: str, os.Pathlike

    :raises: ``SystemExit`` if above checks don't pass.
    """
    match = NEWS_PATTERN.fullmatch(Path(name).name)
    if match is None:
        raise SystemExit(
            f"'{name}' does not match the '(pr-number).(type).rst' or '(pr-number).(type).(enumeration).rst' changelog "
            f"entries formats. See:\n{CHANGELOG_GUIDE}"
        )

    if match.group(2) not in CHANGE_TYPES:
        sys.exit("'{}' of of invalid type '{}'. Valid types are:\n{}".format(name, match.group(2), CHANGE_TYPES))

    print(name, "is ok")


def main():
    for file in NEWS_DIR.iterdir():
        if file.name in ["README.txt", "_template.rst", ".gitignore"]:
            continue
        validate_name(file)


if __name__ == "__main__":
    main()



#-----------------------------------------------------------------------------
# Copyright (c) 2015-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This module, when imported as 'pyi_testmod_metapath1.extern.ccc.ddd', has actually has this __name__, even if the
# parent module's __name__ is 'pyi_testmod_metapath1._vendor.ccc'. This is because the parent module is known to
# sys.modules as '.extern.ccc'.

assert __name__.endswith('.ccc.ddd'), __name__



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.Tools import ccroot, ar
from waflib.Configure import conf


@conf
def find_gxx(conf):
    cxx = conf.find_program(['g++', 'c++'], var='CXX')
    conf.get_cc_version(cxx, gcc=True)
    conf.env.CXX_NAME = 'gcc'


@conf
def gxx_common_flags(conf):
    v = conf.env
    v.CXX_SRC_F = []
    v.CXX_TGT_F = ['-c', '-o']
    if not v.LINK_CXX:
        v.LINK_CXX = v.CXX
    v.CXXLNK_SRC_F = []
    v.CXXLNK_TGT_F = ['-o']
    v.CPPPATH_ST = '-I%s'
    v.DEFINES_ST = '-D%s'
    v.LIB_ST = '-l%s'
    v.LIBPATH_ST = '-L%s'
    v.STLIB_ST = '-l%s'
    v.STLIBPATH_ST = '-L%s'
    v.RPATH_ST = '-Wl,-rpath,%s'
    v.SONAME_ST = '-Wl,-h,%s'
    v.SHLIB_MARKER = '-Wl,-Bdynamic'
    v.STLIB_MARKER = '-Wl,-Bstatic'
    v.cxxprogram_PATTERN = '%s'
    v.CXXFLAGS_cxxshlib = ['-fPIC']
    v.LINKFLAGS_cxxshlib = ['-shared']
    v.cxxshlib_PATTERN = 'lib%s.so'
    v.LINKFLAGS_cxxstlib = ['-Wl,-Bstatic']
    v.cxxstlib_PATTERN = 'lib%s.a'
    v.LINKFLAGS_MACBUNDLE = ['-bundle', '-undefined', 'dynamic_lookup']
    v.CXXFLAGS_MACBUNDLE = ['-fPIC']
    v.macbundle_PATTERN = '%s.bundle'


@conf
def gxx_modifier_win32(conf):
    v = conf.env
    v.cxxprogram_PATTERN = '%s.exe'
    v.cxxshlib_PATTERN = '%s.dll'
    v.implib_PATTERN = '%s.dll.a'
    v.IMPLIB_ST = '-Wl,--out-implib,%s'
    v.CXXFLAGS_cxxshlib = []
    v.append_value('LINKFLAGS', ['-Wl,--enable-auto-import'])


@conf
def gxx_modifier_cygwin(conf):
    gxx_modifier_win32(conf)
    v = conf.env
    v.cxxshlib_PATTERN = 'cyg%s.dll'
    v.append_value('LINKFLAGS_cxxshlib', ['-Wl,--enable-auto-image-base'])
    v.CXXFLAGS_cxxshlib = []


@conf
def gxx_modifier_darwin(conf):
    v = conf.env
    v.CXXFLAGS_cxxshlib = ['-fPIC']
    v.LINKFLAGS_cxxshlib = ['-dynamiclib']
    v.cxxshlib_PATTERN = 'lib%s.dylib'
    v.FRAMEWORKPATH_ST = '-F%s'
    v.FRAMEWORK_ST = ['-framework']
    v.ARCH_ST = ['-arch']
    v.LINKFLAGS_cxxstlib = []
    v.SHLIB_MARKER = []
    v.STLIB_MARKER = []
    v.SONAME_ST = []


@conf
def gxx_modifier_aix(conf):
    v = conf.env
    v.LINKFLAGS_cxxprogram = ['-Wl,-brtl']
    v.LINKFLAGS_cxxshlib = ['-shared', '-Wl,-brtl,-bexpfull']
    v.SHLIB_MARKER = []


@conf
def gxx_modifier_hpux(conf):
    v = conf.env
    v.SHLIB_MARKER = []
    v.STLIB_MARKER = []
    v.CFLAGS_cxxshlib = ['-fPIC', '-DPIC']
    v.cxxshlib_PATTERN = 'lib%s.sl'


@conf
def gxx_modifier_openbsd(conf):
    conf.env.SONAME_ST = []


@conf
def gcc_modifier_osf1V(conf):
    v = conf.env
    v.SHLIB_MARKER = []
    v.STLIB_MARKER = []
    v.SONAME_ST = []


@conf
def gxx_modifier_platform(conf):
    gxx_modifier_func = getattr(conf, 'gxx_modifier_' + conf.env.DEST_OS, None)
    if gxx_modifier_func:
        gxx_modifier_func()


def configure(conf):
    conf.find_gxx()
    conf.find_ar()
    conf.gxx_common_flags()
    conf.gxx_modifier_platform()
    conf.cxx_load_tools()
    conf.cxx_add_flags()
    conf.link_add_flags()
    conf.check_gcc_o_space('cxx')



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Build packages using spec files.

NOTE: All global variables, classes and imported modules create API for .spec files.
"""

import glob
import os
import pathlib
import pprint
import shutil
import enum
import re
import sys

from PyInstaller import DEFAULT_DISTPATH, DEFAULT_WORKPATH, HOMEPATH, compat
from PyInstaller import log as logging
from PyInstaller.building.api import COLLECT, EXE, MERGE, PYZ
from PyInstaller.building.datastruct import (
    TOC, Target, Tree, _check_guts_eq, normalize_toc, normalize_pyz_toc, toc_process_symbolic_links
)
from PyInstaller.building.osx import BUNDLE
from PyInstaller.building.splash import Splash
from PyInstaller.building.utils import (
    _check_guts_toc, _check_guts_toc_mtime, _should_include_system_binary, format_binaries_and_datas, compile_pymodule,
    destination_name_for_extension, postprocess_binaries_toc_pywin32, postprocess_binaries_toc_pywin32_anaconda,
    create_base_library_zip
)
from PyInstaller.compat import is_win, is_conda, is_darwin, is_linux
from PyInstaller.depend import bindepend
from PyInstaller.depend.analysis import initialize_modgraph, HOOK_PRIORITY_USER_HOOKS
from PyInstaller.depend.utils import scan_code_for_ctypes
from PyInstaller import isolated
from PyInstaller.utils.misc import absnormpath, get_path_to_toplevel_modules, mtime
from PyInstaller.utils.hooks import get_package_paths
from PyInstaller.utils.hooks.gi import compile_glib_schema_files

if is_darwin:
    from PyInstaller.utils import osx as osxutils

logger = logging.getLogger(__name__)

STRINGTYPE = type('')
TUPLETYPE = type((None,))

rthooks = {}

# Place where the loader modules and initialization scripts live.
_init_code_path = os.path.join(HOMEPATH, 'PyInstaller', 'loader')

IMPORT_TYPES = [
    'top-level', 'conditional', 'delayed', 'delayed, conditional', 'optional', 'conditional, optional',
    'delayed, optional', 'delayed, conditional, optional'
]

WARNFILE_HEADER = """\

This file lists modules PyInstaller was not able to find. This does not
necessarily mean these modules are required for running your program. Both
Python's standard library and 3rd-party Python packages often conditionally
import optional modules, some of which may be available only on certain
platforms.

Types of import:
* top-level: imported at the top-level - look at these first
* conditional: imported within an if-statement
* delayed: imported within a function
* optional: imported within a try-except-statement

IMPORTANT: Do NOT post this list to the issue-tracker. Use it as a basis for
            tracking down the missing module yourself. Thanks!

"""


@isolated.decorate
def discover_hook_directories():
    """
    Discover hook directories via pyinstaller40 entry points. Perform the discovery in an isolated subprocess
    to avoid importing the package(s) in the main process.

    :return: list of discovered hook directories.
    """

    from traceback import format_exception_only
    from PyInstaller.log import logger
    from PyInstaller.compat import importlib_metadata
    from PyInstaller.depend.analysis import HOOK_PRIORITY_CONTRIBUTED_HOOKS, HOOK_PRIORITY_UPSTREAM_HOOKS

    # The “selectable” entry points (via group and name keyword args) were introduced in importlib_metadata 4.6 and
    # Python 3.10. The compat module ensures we are using a compatible version.
    entry_points = importlib_metadata.entry_points(group='pyinstaller40', name='hook-dirs')

    # Ensure that pyinstaller_hooks_contrib comes last so that hooks from packages providing their own take priority.
    # In pyinstaller-hooks-contrib >= 2024.8, the entry-point module is `_pyinstaller_hooks_contrib`; in earlier
    # versions, it was `_pyinstaller_hooks_contrib.hooks`.
    entry_points = sorted(entry_points, key=lambda x: x.module.startswith("_pyinstaller_hooks_contrib"))

    hook_directories = []
    for entry_point in entry_points:
        # Query hook directory location(s) from entry point
        try:
            hook_directory_entries = entry_point.load()()
        except Exception as e:
            msg = "".join(format_exception_only(type(e), e)).strip()
            logger.warning("discover_hook_directories: Failed to process hook entry point '%s': %s", entry_point, msg)
            continue

        # Determine location-based priority: upstream hooks vs. hooks from contributed hooks package.
        location_priority = (
            HOOK_PRIORITY_CONTRIBUTED_HOOKS
            if entry_point.module.startswith("_pyinstaller_hooks_contrib") else HOOK_PRIORITY_UPSTREAM_HOOKS
        )

        # Append entries
        hook_directories.extend([(hook_directory_entry, location_priority)
                                 for hook_directory_entry in hook_directory_entries])

    logger.debug("discover_hook_directories: Hook directories: %s", hook_directories)

    return hook_directories


def find_binary_dependencies(binaries, import_packages, symlink_suppression_patterns):
    """
    Find dynamic dependencies (linked shared libraries) for the provided list of binaries.

    On Windows, this function performs additional pre-processing in an isolated environment in an attempt to handle
    dynamic library search path modifications made by packages during their import. The packages from the given list
    of collected packages are imported one by one, while keeping track of modifications made by `os.add_dll_directory`
    calls and additions to the `PATH`  environment variable. The recorded additional search paths are then passed to
    the binary dependency analysis step.

    binaries
            List of binaries to scan for dynamic dependencies.
    import_packages
            List of packages to import prior to scanning binaries.
    symlink_suppression_patterns
            Set of paths and/or path patterns for which binary dependency analysis should not create symbolic links
            to the top-level application directory (when the discovered shared library's parent directory structure
            is preserved). When binary dependency analysis discovers a shared library, it matches its *source path*
            against all symlink suppression patterns (using `pathlib.PurePath.match`) to determine whether to create
            a symbolic link to top-level application directory or not.

    :return: expanded list of binaries and then dependencies.
    """

    # Extra library search paths (used on Windows to resolve DLL paths).
    extra_libdirs = []
    if compat.is_win:
        # Always search `sys.base_prefix`, and search it first. This ensures that we resolve the correct version of
        # `python3X.dll` and `python3.dll` (a PEP-0384 stable ABI stub that forwards symbols to the fully versioned
        # `python3X.dll`), regardless of other python installations that might be present in the PATH.
        extra_libdirs.append(compat.base_prefix)

        # When using python built from source, `sys.base_prefix` does not point to the directory that contains python
        # executable, `python3X.dll`, and `python3.dll`. To accommodate such case, also add the directory in which
        # python executable is located to the extra search paths. On the off-chance that this is a combination of venv
        # and python built from source, prefer `sys._base_executable` over `sys.executable`.
        extra_libdirs.append(os.path.dirname(getattr(sys, '_base_executable', sys.executable)))

        # If `pywin32` is installed, resolve the path to the `pywin32_system32` directory. Most `pywin32` extensions
        # reference the `pywintypes3X.dll` in there. Based on resolved `pywin32_system32` directory, also add other
        # `pywin32` directory, in case extensions in different directories reference each other (the ones in the same
        # directory should already be resolvable due to binary dependency analysis passing the analyzed binary's
        # location to the `get_imports` function). This allows us to avoid searching all paths in `sys.path`, which
        # may lead to other corner-case issues (e.g., #5560).
        pywin32_system32_dir = None
        try:
            # Look up the directory by treating it as a namespace package.
            _, pywin32_system32_dir = get_package_paths('pywin32_system32')
        except Exception:
            pass

        if pywin32_system32_dir:
            pywin32_base_dir = os.path.dirname(pywin32_system32_dir)
            extra_libdirs += [
                pywin32_system32_dir,  # .../pywin32_system32
                # based on pywin32.pth
                os.path.join(pywin32_base_dir, 'win32'),  # .../win32
                os.path.join(pywin32_base_dir, 'win32', 'lib'),  # .../win32/lib
                os.path.join(pywin32_base_dir, 'Pythonwin'),  # .../Pythonwin
            ]

    # On Windows, packages' initialization code might register additional DLL search paths, either by modifying the
    # `PATH` environment variable, or by calling `os.add_dll_directory`. Therefore, we import all collected packages,
    # and track changes made to the environment.
    if compat.is_win:
        # Helper functions to be executed in isolated environment.
        def setup(suppressed_imports):
            """
            Prepare environment for change tracking
            """
            import os
            import sys

            os._added_dll_directories = []
            os._original_path_env = os.environ.get('PATH', '')

            _original_add_dll_directory = os.add_dll_directory

            def _pyi_add_dll_directory(path):
                os._added_dll_directories.append(path)
                return _original_add_dll_directory(path)

            os.add_dll_directory = _pyi_add_dll_directory

            # Suppress import of specified packages
            for name in suppressed_imports:
                sys.modules[name] = None

        def import_library(package):
            """
            Import collected package to set up environment.
            """
            try:
                __import__(package)
            except Exception:
                pass

        def process_search_paths():
            """
            Obtain lists of added search paths.
            """
            import os

            # `os.add_dll_directory` might be called with a `pathlib.Path`, which cannot be marhsalled out of the helper
            # process. So explicitly convert all entries to strings.
            dll_directories = [str(path) for path in os._added_dll_directories]

            orig_path = set(os._original_path_env.split(os.pathsep))
            modified_path = os.environ.get('PATH', '').split(os.pathsep)
            path_additions = [path for path in modified_path if path and path not in orig_path]

            return dll_directories, path_additions

        # Pre-process the list of packages to import.
        # Check for Qt bindings packages, and put them at the front of the packages list. This ensures that they are
        # always imported first, which should prevent packages that support multiple bindings (`qtypy`, `pyqtgraph`,
        # `matplotlib`, etc.) from trying to auto-select bindings.
        _QT_BINDINGS = ('PySide2', 'PyQt5', 'PySide6', 'PyQt6')

        qt_packages = []
        other_packages = []
        for package in import_packages:
            if package.startswith(_QT_BINDINGS):
                qt_packages.append(package)
            else:
                other_packages.append(package)
        import_packages = qt_packages + other_packages

        # Just in case, explicitly suppress imports of Qt bindings that we are *not* collecting - if multiple bindings
        # are available and some were excluded from our analysis, a package imported here might still try to import an
        # excluded bindings package (and succeed at doing so).
        suppressed_imports = [package for package in _QT_BINDINGS if package not in qt_packages]

        # If we suppressed PySide2 or PySide6, we must also suppress their corresponding shiboken package
        if "PySide2" in suppressed_imports:
            suppressed_imports += ["shiboken2"]
        if "PySide6" in suppressed_imports:
            suppressed_imports += ["shiboken6"]

        # Suppress import of `pyqtgraph.canvas`, which is known to crash python interpreter. See #7452 and #8322, as
        # well as https://github.com/pyqtgraph/pyqtgraph/issues/2838.
        suppressed_imports += ['pyqtgraph.canvas']

        # PySimpleGUI 5.x displays a "first-run" dialog when imported for the first time, which blocks the loop below.
        # This is a problem for building on CI, where the dialog cannot be closed, and where PySimpleGUI runs "for the
        # first time" every time. See #8396.
        suppressed_imports += ['PySimpleGUI']

        # Processing in isolated environment.
        with isolated.Python() as child:
            child.call(setup, suppressed_imports)
            for package in import_packages:
                try:
                    child.call(import_library, package)
                except isolated.SubprocessDiedError as e:
                    # Re-raise as `isolated.SubprocessDiedError` again, to trigger error-handling codepath in
                    # `isolated.Python.__exit__()`.
                    raise isolated.SubprocessDiedError(
                        f"Isolated subprocess crashed while importing package {package!r}! "
                        f"Package import list: {import_packages!r}"
                    ) from e
            added_dll_directories, added_path_directories = child.call(process_search_paths)

        # Process extra search paths...
        logger.info("Extra DLL search directories (AddDllDirectory): %r", added_dll_directories)
        extra_libdirs += added_dll_directories

        logger.info("Extra DLL search directories (PATH): %r", added_path_directories)
        extra_libdirs += added_path_directories

    # Deduplicate search paths
    # NOTE: `list(set(extra_libdirs))` does not preserve the order of search paths (which matters here), so we need to
    # de-duplicate using `list(dict.fromkeys(extra_libdirs).keys())` instead.
    extra_libdirs = list(dict.fromkeys(extra_libdirs).keys())

    # Search for dependencies of the given binaries
    return bindepend.binary_dependency_analysis(
        binaries,
        search_paths=extra_libdirs,
        symlink_suppression_patterns=symlink_suppression_patterns,
    )


class _ModuleCollectionMode(enum.IntFlag):
    """
    Module collection mode flags.
    """
    PYZ = enum.auto()  # Collect byte-compiled .pyc into PYZ archive
    PYC = enum.auto()  # Collect byte-compiled .pyc as external data file
    PY = enum.auto()  # Collect source .py file as external data file


_MODULE_COLLECTION_MODES = {
    "pyz": _ModuleCollectionMode.PYZ,
    "pyc": _ModuleCollectionMode.PYC,
    "py": _ModuleCollectionMode.PY,
    "pyz+py": _ModuleCollectionMode.PYZ | _ModuleCollectionMode.PY,
    "py+pyz": _ModuleCollectionMode.PYZ | _ModuleCollectionMode.PY,
}


def _get_module_collection_mode(mode_dict, name, noarchive=False):
    """
    Determine the module/package collection mode for the given module name, based on the provided collection
    mode settings dictionary.
    """
    # Default mode: collect into PYZ, unless noarchive is enabled. In that case, collect as pyc.
    mode_flags = _ModuleCollectionMode.PYC if noarchive else _ModuleCollectionMode.PYZ

    # If we have no collection mode settings, end here and now.
    if not mode_dict:
        return mode_flags

    # Search the parent modules/packages in top-down fashion, and take the last given setting. This ensures that
    # a setting given for the top-level package is recursively propagated to all its subpackages and submodules,
    # but also allows individual sub-modules to override the setting again.
    mode = 'pyz'

    name_parts = name.split('.')
    for i in range(len(name_parts)):
        modlevel = ".".join(name_parts[:i + 1])
        modlevel_mode = mode_dict.get(modlevel, None)
        if modlevel_mode is not None:
            mode = modlevel_mode

    # Convert mode string to _ModuleCollectionMode flags
    try:
        mode_flags = _MODULE_COLLECTION_MODES[mode]
    except KeyError:
        raise ValueError(f"Unknown module collection mode for {name!r}: {mode!r}!")

    # noarchive flag being set means that we need to change _ModuleCollectionMode.PYZ into _ModuleCollectionMode.PYC
    if noarchive and _ModuleCollectionMode.PYZ in mode_flags:
        mode_flags ^= _ModuleCollectionMode.PYZ
        mode_flags |= _ModuleCollectionMode.PYC

    return mode_flags


class Analysis(Target):
    """
    Class that performs analysis of the user's main Python scripts.

    An Analysis contains multiple TOC (Table of Contents) lists, accessed as attributes of the analysis object.

    scripts
            The scripts you gave Analysis as input, with any runtime hook scripts prepended.
    pure
            The pure Python modules.
    binaries
            The extension modules and their dependencies.
    datas
            Data files collected from packages.
    zipfiles
            Deprecated - always empty.
    zipped_data
            Deprecated - always empty.
    """
    _old_scripts = {
        absnormpath(os.path.join(HOMEPATH, "support", "_mountzlib.py")),
        absnormpath(os.path.join(HOMEPATH, "support", "useUnicode.py")),
        absnormpath(os.path.join(HOMEPATH, "support", "useTK.py")),
        absnormpath(os.path.join(HOMEPATH, "support", "unpackTK.py")),
        absnormpath(os.path.join(HOMEPATH, "support", "removeTK.py"))
    }

    def __init__(
        self,
        scripts,
        pathex=None,
        binaries=None,
        datas=None,
        hiddenimports=None,
        hookspath=None,
        hooksconfig=None,
        excludes=None,
        runtime_hooks=None,
        cipher=None,
        win_no_prefer_redirects=False,
        win_private_assemblies=False,
        noarchive=False,
        module_collection_mode=None,
        optimize=-1,
        **_kwargs,
    ):
        """
        scripts
                A list of scripts specified as file names.
        pathex
                An optional list of paths to be searched before sys.path.
        binaries
                An optional list of additional binaries (dlls, etc.) to include.
        datas
                An optional list of additional data files to include.
        hiddenimport
                An optional list of additional (hidden) modules to include.
        hookspath
                An optional list of additional paths to search for hooks. (hook-modules).
        hooksconfig
                An optional dict of config settings for hooks. (hook-modules).
        excludes
                An optional list of module or package names (their Python names, not path names) that will be
                ignored (as though they were not found).
        runtime_hooks
                An optional list of scripts to use as users' runtime hooks. Specified as file names.
        cipher
                Deprecated. Raises an error if not None.
        win_no_prefer_redirects
                Deprecated. Raises an error if not False.
        win_private_assemblies
                Deprecated. Raises an error if not False.
        noarchive
                If True, do not place source files in a archive, but keep them as individual files.
        module_collection_mode
                An optional dict of package/module names and collection mode strings. Valid collection mode strings:
                'pyz' (default), 'pyc', 'py', 'pyz+py' (or 'py+pyz')
        optimize
                Optimization level for collected bytecode. If not specified or set to -1, it is set to the value of
                `sys.flags.optimize` of the running build process.
        """
        if cipher is not None:
            from PyInstaller.exceptions import RemovedCipherFeatureError
            raise RemovedCipherFeatureError(
                "Please remove the 'cipher' arguments to PYZ() and Analysis() in your spec file."
            )
        if win_no_prefer_redirects:
            from PyInstaller.exceptions import RemovedWinSideBySideSupportError
            raise RemovedWinSideBySideSupportError(
                "Please remove the 'win_no_prefer_redirects' argument to Analysis() in your spec file."
            )
        if win_private_assemblies:
            from PyInstaller.exceptions import RemovedWinSideBySideSupportError
            raise RemovedWinSideBySideSupportError(
                "Please remove the 'win_private_assemblies' argument to Analysis() in your spec file."
            )
        super().__init__()
        from PyInstaller.config import CONF

        self.inputs = []
        spec_dir = os.path.dirname(CONF['spec'])
        for script in scripts:
            # If path is relative, it is relative to the location of .spec file.
            if not os.path.isabs(script):
                script = os.path.join(spec_dir, script)
            if absnormpath(script) in self._old_scripts:
                logger.warning('Ignoring obsolete auto-added script %s', script)
                continue
            # Normalize script path.
            script = os.path.normpath(script)
            if not os.path.exists(script):
                raise SystemExit("ERROR: script '%s' not found" % script)
            self.inputs.append(script)

        # Django hook requires this variable to find the script manage.py.
        CONF['main_script'] = self.inputs[0]

        site_packages_pathex = []
        for path in (pathex or []):
            if pathlib.Path(path).name == "site-packages":
                site_packages_pathex.append(str(path))
        if site_packages_pathex:
            logger.log(
                logging.DEPRECATION, "Foreign Python environment's site-packages paths added to --paths/pathex:\n%s\n"
                "This is ALWAYS the wrong thing to do. If your environment's site-packages is not in PyInstaller's "
                "module search path then you are running PyInstaller from a different environment to the one your "
                "packages are in. Run print(sys.prefix) without PyInstaller to get the environment you should be using "
                "then install and run PyInstaller from that environment instead of this one. This warning will become "
                "an error in PyInstaller 7.0.", pprint.pformat(site_packages_pathex)
            )

        self.pathex = self._extend_pathex(pathex, self.inputs)
        # Set global config variable 'pathex' to make it available for PyInstaller.utils.hooks and import hooks. Path
        # extensions for module search.
        CONF['pathex'] = self.pathex
        # Extend sys.path so PyInstaller could find all necessary modules.
        sys.path.extend(self.pathex)
        logger.info('Module search paths (PYTHONPATH):\n' + pprint.pformat(sys.path))

        self.hiddenimports = hiddenimports or []
        # Include hidden imports passed via CONF['hiddenimports']; these might be populated if user has a wrapper script
        # that calls `build_main.main()` with custom `pyi_config` dictionary that contains `hiddenimports`.
        self.hiddenimports.extend(CONF.get('hiddenimports', []))

        for modnm in self.hiddenimports:
            if re.search(r"[\\/]", modnm):
                raise SystemExit(
                    f"ERROR: Invalid hiddenimport '{modnm}'. Hidden imports should be importable module names – not "
                    "file paths. i.e. use --hiddenimport=foo.bar instead of --hiddenimport=.../site-packages/foo/bar.py"
                )

        self.hookspath = []
        # Prepend directories in `hookspath` (`--additional-hooks-dir`) to take precedence over those from the entry
        # points. Expand starting tilde into user's home directory, as a work-around for tilde not being expanded by
        # shell when using `--additional-hooks-dir=~/path/abc` instead of `--additional-hooks-dir ~/path/abc` (or when
        # the path argument is quoted).
        if hookspath:
            self.hookspath.extend([(os.path.expanduser(path), HOOK_PRIORITY_USER_HOOKS) for path in hookspath])

        # Add hook directories from PyInstaller entry points.
        self.hookspath += discover_hook_directories()

        self.hooksconfig = {}
        if hooksconfig:
            self.hooksconfig.update(hooksconfig)

        # Custom runtime hook files that should be included and started before any existing PyInstaller runtime hooks.
        self.custom_runtime_hooks = runtime_hooks or []

        self._input_binaries = []
        self._input_datas = []

        self.excludes = excludes or []
        self.scripts = []
        self.pure = []
        self.binaries = []
        self.zipfiles = []
        self.zipped_data = []
        self.datas = []
        self.dependencies = []
        self._python_version = sys.version
        self.noarchive = noarchive
        self.module_collection_mode = module_collection_mode or {}
        self.optimize = sys.flags.optimize if optimize in {-1, None} else optimize

        self._modules_outside_pyz = []

        # Validate the optimization level to avoid errors later on...
        if self.optimize not in {0, 1, 2}:
            raise ValueError(f"Unsupported bytecode optimization level: {self.optimize!r}")

        # Expand the `binaries` and `datas` lists specified in the .spec file, and ensure that the lists are normalized
        # and sorted before guts comparison.
        #
        # While we use these lists to initialize `Analysis.binaries` and `Analysis.datas`, at this point, we need to
        # store them in separate variables, which undergo *full* guts comparison (`_check_guts_toc`) as opposed to
        # just mtime-based comparison (`_check_guts_toc_mtime`). Changes to these initial list *must* trigger a rebuild
        # (and due to the way things work, a re-analysis), otherwise user might end up with a cached build that fails to
        # reflect the changes.
        if binaries:
            logger.info("Appending 'binaries' from .spec")
            self._input_binaries = [(dest_name, src_name, 'BINARY')
                                    for dest_name, src_name in format_binaries_and_datas(binaries, workingdir=spec_dir)]
            self._input_binaries = sorted(normalize_toc(self._input_binaries))

        if datas:
            logger.info("Appending 'datas' from .spec")
            self._input_datas = [(dest_name, src_name, 'DATA')
                                 for dest_name, src_name in format_binaries_and_datas(datas, workingdir=spec_dir)]
            self._input_datas = sorted(normalize_toc(self._input_datas))

        self.__postinit__()

    _GUTS = (  # input parameters
        ('inputs', _check_guts_eq),  # parameter `scripts`
        ('pathex', _check_guts_eq),
        ('hiddenimports', _check_guts_eq),
        ('hookspath', _check_guts_eq),
        ('hooksconfig', _check_guts_eq),
        ('excludes', _check_guts_eq),
        ('custom_runtime_hooks', _check_guts_eq),
        ('noarchive', _check_guts_eq),
        ('module_collection_mode', _check_guts_eq),
        ('optimize', _check_guts_eq),

        ('_input_binaries', _check_guts_toc),
        ('_input_datas', _check_guts_toc),

        # calculated/analysed values
        ('_python_version', _check_guts_eq),
        ('scripts', _check_guts_toc_mtime),
        ('pure', _check_guts_toc_mtime),
        ('binaries', _check_guts_toc_mtime),
        ('zipfiles', _check_guts_toc_mtime),
        ('zipped_data', None),  # TODO check this, too
        ('datas', _check_guts_toc_mtime),
        # TODO: Need to add "dependencies"?

        ('_modules_outside_pyz', _check_guts_toc_mtime),
    )

    def _extend_pathex(self, spec_pathex, scripts):
        """
        Normalize additional paths where PyInstaller will look for modules and add paths with scripts to the list of
        paths.

        :param spec_pathex: Additional paths defined defined in .spec file.
        :param scripts: Scripts to create executable from.
        :return: list of updated paths
        """
        # Based on main supplied script - add top-level modules directory to PYTHONPATH.
        # Sometimes the main app script is not top-level module but submodule like 'mymodule.mainscript.py'.
        # In that case PyInstaller will not be able find modules in the directory containing 'mymodule'.
        # Add this directory to PYTHONPATH so PyInstaller could find it.
        pathex = []
        # Add scripts paths first.
        for script in scripts:
            logger.debug('script: %s' % script)
            script_toplevel_dir = get_path_to_toplevel_modules(script)
            if script_toplevel_dir:
                pathex.append(script_toplevel_dir)
        # Append paths from .spec.
        if spec_pathex is not None:
            pathex.extend(spec_pathex)
        # Normalize paths in pathex and make them absolute.
        return list(dict.fromkeys(absnormpath(p) for p in pathex))

    def _check_guts(self, data, last_build):
        if Target._check_guts(self, data, last_build):
            return True
        for filename in self.inputs:
            if mtime(filename) > last_build:
                logger.info("Building because %s changed", filename)
                return True
        # Now we know that none of the input parameters and none of the input files has changed. So take the values
        # that were calculated / analyzed in the last run and store them in `self`. These TOC lists should already
        # be normalized.
        self.scripts = data['scripts']
        self.pure = data['pure']
        self.binaries = data['binaries']
        self.zipfiles = data['zipfiles']
        self.zipped_data = data['zipped_data']
        self.datas = data['datas']

        return False

    def assemble(self):
        """
        This method is the MAIN method for finding all necessary files to be bundled.
        """
        from PyInstaller.config import CONF

        # Search for python shared library, which we need to collect into frozen application. Do this as the very
        # first step, to minimize the amount of processing when the shared library cannot be found.
        logger.info('Looking for Python shared library...')
        python_lib = bindepend.get_python_library_path()  # Raises PythonLibraryNotFoundError
        logger.info('Using Python shared library: %s', python_lib)

        logger.info("Running Analysis %s", self.tocbasename)
        logger.info("Target bytecode optimization level: %d", self.optimize)

        for m in self.excludes:
            logger.debug("Excluding module '%s'" % m)
        self.graph = initialize_modgraph(excludes=self.excludes, user_hook_dirs=self.hookspath)

        # Initialize `binaries` and `datas` with `_input_binaries` and `_input_datas`. Make sure to copy the lists
        # to prevent modifications of original lists, which we need to store in original form for guts comparison.
        self.datas = [entry for entry in self._input_datas]
        self.binaries = [entry for entry in self._input_binaries]

        # Expand sys.path of module graph. The attribute is the set of paths to use for imports: sys.path, plus our
        # loader, plus other paths from e.g. --path option).
        self.graph.path = self.pathex + self.graph.path

        # Scan for legacy namespace packages.
        self.graph.scan_legacy_namespace_packages()

        # Add python shared library to `binaries`.
        if is_darwin and osxutils.is_framework_bundle_lib(python_lib):
            # If python library is located in macOS .framework bundle, collect the bundle, and create symbolic link to
            # top-level directory.
            src_path = pathlib.PurePath(python_lib)
            dst_path = pathlib.PurePath(src_path.relative_to(src_path.parent.parent.parent.parent))
            self.binaries.append((str(dst_path), str(src_path), 'BINARY'))
            self.binaries.append((os.path.basename(python_lib), str(dst_path), 'SYMLINK'))
        else:
            self.binaries.append((os.path.basename(python_lib), python_lib, 'BINARY'))

        # -- Module graph. --
        #
        # Construct the module graph of import relationships between modules required by this user's application. For
        # each entry point (top-level user-defined Python script), all imports originating from this entry point are
        # recursively parsed into a subgraph of the module graph. This subgraph is then connected to this graph's root
        # node, ensuring imported module nodes will be reachable from the root node -- which is is (arbitrarily) chosen
        # to be the first entry point's node.

        # List of graph nodes corresponding to program scripts.
        program_scripts = []

        # Assume that if the script does not exist, Modulegraph will raise error. Save the graph nodes of each in
        # sequence.
        for script in self.inputs:
            logger.info("Analyzing %s", script)
            program_scripts.append(self.graph.add_script(script))

        # Analyze the script's hidden imports (named on the command line)
        self.graph.add_hiddenimports(self.hiddenimports)

        # -- Post-graph hooks. --
        self.graph.process_post_graph_hooks(self)

        # Update 'binaries' and 'datas' TOC lists with entries collected from hooks.
        self.binaries += self.graph.make_hook_binaries_toc()
        self.datas += self.graph.make_hook_datas_toc()

        # We do not support zipped eggs anymore (PyInstaller v6.0), so `zipped_data` and `zipfiles` are always empty.
        self.zipped_data = []
        self.zipfiles = []

        # -- Automatic binary vs. data reclassification. --
        #
        # At this point, `binaries` and `datas` contain  TOC entries supplied by user via input arguments, and by hooks
        # that were ran during the analysis. Neither source can be fully trusted regarding the DATA vs BINARY
        # classification (no thanks to our hookutils not being 100% reliable, either!). Therefore, inspect the files and
        # automatically reclassify them as necessary.
        #
        # The proper classification is important especially for collected binaries - to ensure that they undergo binary
        # dependency analysis and platform-specific binary processing. On macOS, the .app bundle generation code also
        # depends on files to be properly classified.
        #
        # For entries added to `binaries` and `datas` after this point, we trust their typecodes due to the nature of
        # their origin.
        combined_toc = normalize_toc(self.datas + self.binaries)

        logger.info('Performing binary vs. data reclassification (%d entries)', len(combined_toc))

        self.datas = []
        self.binaries = []

        for dest_name, src_name, typecode in combined_toc:
            # Returns 'BINARY' or 'DATA', or None if file cannot be classified.
            detected_typecode = bindepend.classify_binary_vs_data(src_name)
            if detected_typecode is not None:
                if detected_typecode != typecode:
                    logger.debug(
                        "Reclassifying collected file %r from %s to %s...", src_name, typecode, detected_typecode
                    )
                typecode = detected_typecode

            # Put back into corresponding TOC list.
            if typecode in {'BINARY', 'EXTENSION'}:
                self.binaries.append((dest_name, src_name, typecode))
            else:
                self.datas.append((dest_name, src_name, typecode))

        # -- Look for dlls that are imported by Python 'ctypes' module. --
        # First get code objects of all modules that import 'ctypes'.
        logger.info('Looking for ctypes DLLs')
        # dict like: {'module1': code_obj, 'module2': code_obj}
        ctypes_code_objs = self.graph.get_code_using("ctypes")

        for name, co in ctypes_code_objs.items():
            # Get dlls that might be needed by ctypes.
            logger.debug('Scanning %s for ctypes-based references to shared libraries', name)
            try:
                ctypes_binaries = scan_code_for_ctypes(co)
                # As this scan happens after automatic binary-vs-data classification, we need to validate the binaries
                # ourselves, just in case.
                for dest_name, src_name, typecode in set(ctypes_binaries):
                    # Allow for `None` in case re-classification is not supported on the given platform.
                    if bindepend.classify_binary_vs_data(src_name) not in (None, 'BINARY'):
                        logger.warning("Ignoring %s found via ctypes - not a valid binary!", src_name)
                        continue
                    self.binaries.append((dest_name, src_name, typecode))
            except Exception as ex:
                raise RuntimeError(f"Failed to scan the module '{name}'. This is a bug. Please report it.") from ex

        self.datas.extend((dest, source, "DATA")
                          for (dest, source) in format_binaries_and_datas(self.graph.metadata_required()))

        # Analyze run-time hooks.
        rhtook_scripts = self.graph.analyze_runtime_hooks(self.custom_runtime_hooks)

        # -- Extract the nodes of the graph as TOCs for further processing. --

        # Initialize the scripts list: run-time hooks (custom ones, followed by regular ones), followed by program
        # script(s).

        # We do not optimize bytecode of run-time hooks.
        rthook_toc = self.graph.nodes_to_toc(rhtook_scripts)

        # Override the typecode of program script(s) to include bytecode optimization level.
        program_toc = self.graph.nodes_to_toc(program_scripts)
        optim_typecode = {0: 'PYSOURCE', 1: 'PYSOURCE-1', 2: 'PYSOURCE-2'}[self.optimize]
        program_toc = [(name, src_path, optim_typecode) for name, src_path, typecode in program_toc]

        self.scripts = rthook_toc + program_toc
        self.scripts = normalize_toc(self.scripts)  # Should not really contain duplicates, but just in case...

        # Extend the binaries list with all the Extensions modulegraph has found.
        self.binaries += self.graph.make_binaries_toc()

        # Convert extension module names into full filenames, and append suffix. Ensure that extensions that come from
        # the lib-dynload are collected into _MEIPASS/python3.x/lib-dynload instead of directly into _MEIPASS.
        for idx, (dest, source, typecode) in enumerate(self.binaries):
            if typecode != 'EXTENSION':
                continue
            dest = destination_name_for_extension(dest, source, typecode)
            self.binaries[idx] = (dest, source, typecode)

        # Perform initial normalization of `datas` and `binaries`
        self.datas = normalize_toc(self.datas)
        self.binaries = normalize_toc(self.binaries)

        # Post-process GLib schemas
        self.datas = compile_glib_schema_files(self.datas, os.path.join(CONF['workpath'], "_pyi_gschema_compilation"))
        self.datas = normalize_toc(self.datas)

        # Process the pure-python modules list. Depending on the collection mode, these entries end up either in "pure"
        # list for collection into the PYZ archive, or in the "datas" list for collection as external data files.
        assert len(self.pure) == 0
        pure_pymodules_toc = self.graph.make_pure_toc()

        # Merge package collection mode settings from .spec file. These are applied last, so they override the
        # settings previously applied by hooks.
        self.graph._module_collection_mode.update(self.module_collection_mode)
        logger.debug("Module collection settings: %r", self.graph._module_collection_mode)

        # If target bytecode optimization level matches the run-time bytecode optimization level (i.e., of the running
        # build process), we can re-use the modulegraph's code-object cache.
        if self.optimize == sys.flags.optimize:
            logger.debug(
                "Target optimization level %d matches run-time optimization level %d - using modulegraph's code-object "
                "cache.",
                self.optimize,
                sys.flags.optimize,
            )
            code_cache = self.graph.get_code_objects()
        else:
            logger.debug(
                "Target optimization level %d differs from run-time optimization level %d - ignoring modulegraph's "
                "code-object cache.",
                self.optimize,
                sys.flags.optimize,
            )
            code_cache = None

        # Construct a set for look-up of modules that should end up in base_library.zip. The list of corresponding
        # modulegraph nodes is stored in `PyiModuleGraph._base_modules` (see `PyiModuleGraph._analyze_base_modules`).
        base_modules = set(node.identifier for node in self.graph._base_modules)
        base_modules_toc = []

        pycs_dir = os.path.join(CONF['workpath'], 'localpycs')
        optim_level = self.optimize  # We could extend this with per-module settings, similar to `collect_mode`.
        for name, src_path, typecode in pure_pymodules_toc:
            assert typecode == 'PYMODULE'
            collect_mode = _get_module_collection_mode(self.graph._module_collection_mode, name, self.noarchive)

            # Collect byte-compiled .pyc into PYZ archive or base_library.zip. Embed optimization level into typecode.
            in_pyz = False
            if _ModuleCollectionMode.PYZ in collect_mode:
                optim_typecode = {0: 'PYMODULE', 1: 'PYMODULE-1', 2: 'PYMODULE-2'}[optim_level]
                toc_entry = (name, src_path, optim_typecode)
                if name in base_modules:
                    base_modules_toc.append(toc_entry)
                else:
                    self.pure.append(toc_entry)
                    in_pyz = True

            # If module is not collected into PYZ archive (and is consequently not tracked in the `self.pure` TOC list),
            # add it to the `self._modules_outside_pyz` TOC list, in order to be able to detect modifications in those
            # modules.
            if not in_pyz:
                self._modules_outside_pyz.append((name, src_path, typecode))

            # Pure namespace packages have no source path, and cannot be collected as external data file.
            if src_path in (None, '-'):
                continue

            # Collect source .py file as external data file
            if _ModuleCollectionMode.PY in collect_mode:
                basename, ext = os.path.splitext(os.path.basename(src_path))
                # If the module is available only as a byte-compiled .pyc, we cannot collect its source.
                if ext.lower() == '.pyc':
                    logger.warning(
                        'Cannot collect source .py file for module %r - module is available only as .pyc: %r',
                        name,
                        src_path,
                    )
                    continue
                dest_path = name.replace('.', os.sep)
                if basename == '__init__':
                    dest_path += os.sep + '__init__' + ext
                else:
                    dest_path += ext
                self.datas.append((dest_path, src_path, "DATA"))

            # Collect byte-compiled .pyc file as external data file
            if _ModuleCollectionMode.PYC in collect_mode:
                basename, ext = os.path.splitext(os.path.basename(src_path))
                dest_path = name.replace('.', os.sep)
                if basename == '__init__':
                    dest_path += os.sep + '__init__'
                # Append the extension for the compiled result. In python 3.5 (PEP-488) .pyo files were replaced by
                # .opt-1.pyc and .opt-2.pyc. However, it seems that for bytecode-only module distribution, we always
                # need to use the .pyc extension.
                dest_path += '.pyc'

                # Compile - use optimization-level-specific sub-directory in local working directory.
                obj_path = compile_pymodule(
                    name,
                    src_path,
                    workpath=os.path.join(pycs_dir, str(optim_level)),
                    optimize=optim_level,
                    code_cache=code_cache,
                )

                self.datas.append((dest_path, obj_path, "DATA"))

        # Construct base_library.zip, if applicable (the only scenario where it is not is if we are building with
        # noarchive mode). Always remove the file before the build.
        base_library_zip = os.path.join(CONF['workpath'], 'base_library.zip')
        if os.path.exists(base_library_zip):
            os.remove(base_library_zip)
        if base_modules_toc:
            logger.info('Creating %s...', os.path.basename(base_library_zip))
            create_base_library_zip(base_library_zip, base_modules_toc, code_cache)
            self.datas.append((os.path.basename(base_library_zip), base_library_zip, 'DATA'))  # Bundle as data file.

        # Normalize list of pure-python modules (these will end up in PYZ archive, so use specific normalization).
        self.pure = normalize_pyz_toc(self.pure)

        # Associate the `pure` TOC list instance with code cache in the global `CONF`; this is used by `PYZ` writer
        # to obtain modules' code from cache instead
        #
        # (NOTE: back when `pure` was an instance of `TOC` class, the code object was passed by adding an attribute
        # to the `pure` itself; now that `pure` is plain `list`, we cannot do that anymore. But the association via
        # object ID should have the same semantics as the added attribute).
        from PyInstaller.config import CONF
        global_code_cache_map = CONF['code_cache']
        global_code_cache_map[id(self.pure)] = code_cache

        # Add remaining binary dependencies - analyze Python C-extensions and what DLLs they depend on.
        #
        # Up until this point, we did very best not to import the packages into the main process. However, a package
        # may set up additional library search paths during its import (e.g., by modifying PATH or calling the
        # add_dll_directory() function on Windows, or modifying LD_LIBRARY_PATH on Linux). In order to reliably
        # discover dynamic libraries, we therefore require an environment with all packages imported. We achieve that
        # by gathering list of all collected packages, and spawn an isolated process, in which we first import all
        # the packages from the list, and then perform search for dynamic libraries.
        logger.info('Looking for dynamic libraries')

        collected_packages = self.graph.get_collected_packages()
        self.binaries.extend(
            find_binary_dependencies(self.binaries, collected_packages, self.graph._bindepend_symlink_suppression)
        )

        # Apply work-around for (potential) binaries collected from `pywin32` package...
        if is_win:
            self.binaries = postprocess_binaries_toc_pywin32(self.binaries)
            # With anaconda, we need additional work-around...
            if is_conda:
                self.binaries = postprocess_binaries_toc_pywin32_anaconda(self.binaries)

        # On linux, check for HMAC files accompanying shared library files and, if available, collect them.
        # These are present on Fedora and RHEL, and are used in FIPS-enabled configurations to ensure shared
        # library's file integrity.
        if is_linux:
            for dest_name, src_name, typecode in self.binaries:
                if typecode not in {'BINARY', 'EXTENSION'}:
                    continue  # Skip symbolic links

                src_lib_path = pathlib.Path(src_name)

                # Check for .name.hmac file next to the shared library.
                src_hmac_path = src_lib_path.with_name(f".{src_lib_path.name}.hmac")
                if src_hmac_path.is_file():
                    dest_hmac_path = pathlib.PurePath(dest_name).with_name(src_hmac_path.name)
                    self.datas.append((str(dest_hmac_path), str(src_hmac_path), 'DATA'))

                # Alternatively, check the fipscheck directory: fipscheck/name.hmac
                src_hmac_path = src_lib_path.parent / "fipscheck" / f"{src_lib_path.name}.hmac"
                if src_hmac_path.is_file():
                    dest_hmac_path = pathlib.PurePath("fipscheck") / src_hmac_path.name
                    self.datas.append((str(dest_hmac_path), str(src_hmac_path), 'DATA'))

                # Similarly, look for .chk files that are used by NSS libraries.
                src_chk_path = src_lib_path.with_suffix(".chk")
                if src_chk_path.is_file():
                    dest_chk_path = pathlib.PurePath(dest_name).with_name(src_chk_path.name)
                    self.datas.append((str(dest_chk_path), str(src_chk_path), 'DATA'))

        # Final normalization of `datas` and `binaries`:
        #  - normalize both TOCs together (to avoid having duplicates across the lists)
        #  - process the combined normalized TOC for symlinks
        #  - split back into `binaries` (BINARY, EXTENSION) and `datas` (everything else)
        combined_toc = normalize_toc(self.datas + self.binaries)
        combined_toc = toc_process_symbolic_links(combined_toc)

        # On macOS, look for binaries collected from .framework bundles, collect their Info.plist files, and fix the
        # structure to conform to code-signing requirements (i.e., Versions/Current symbolic link and symbolic links
        # for top-level directories).
        if is_darwin:
            combined_toc = osxutils.collect_files_from_framework_bundles(combined_toc)
            combined_toc = normalize_toc(combined_toc)

        self.datas = []
        self.binaries = []
        for entry in combined_toc:
            dest_name, src_name, typecode = entry
            if typecode in {'BINARY', 'EXTENSION'}:
                self.binaries.append(entry)
            else:
                self.datas.append(entry)

        # On macOS, the Finder app seems to litter visited directories with `.DS_Store` files. These cause issues with
        # codesigning when placed in mixed-content directories, where our .app bundle generator cross-links data files
        # from `Resources` to `Frameworks` tree, and the `codesign` utility explicitly forbids a `.DS_Store` file to be
        # a symbolic link.
        # But there is no reason for `.DS_Store` files to be collected in the first place, so filter them out.
        if is_darwin:
            self.datas = [(dest_name, src_name, typecode) for dest_name, src_name, typecode in self.datas
                          if os.path.basename(src_name) != '.DS_Store']

        # Write warnings about missing modules.
        self._write_warnings()
        # Write debug information about the graph
        self._write_graph_debug()

        # On macOS, check the SDK version of the binaries to be collected, and warn when the SDK version is either
        # invalid or too low. Such binaries will likely refuse to be loaded when hardened runtime is enabled and
        # while we cannot do anything about it, we can at least warn the user about it.
        # See: https://developer.apple.com/forums/thread/132526
        if is_darwin:
            binaries_with_invalid_sdk = []
            for dest_name, src_name, typecode in self.binaries:
                try:
                    sdk_version = osxutils.get_macos_sdk_version(src_name)
                except Exception:
                    logger.warning("Failed to query macOS SDK version of %r!", src_name, exc_info=True)
                    binaries_with_invalid_sdk.append((dest_name, src_name, "unavailable"))
                    continue

                if sdk_version < (10, 9, 0):
                    binaries_with_invalid_sdk.append((dest_name, src_name, sdk_version))
            if binaries_with_invalid_sdk:
                logger.warning("Found one or more binaries with invalid or incompatible macOS SDK version:")
                for dest_name, src_name, sdk_version in binaries_with_invalid_sdk:
                    logger.warning(" * %r, collected as %r; version: %r", src_name, dest_name, sdk_version)
                logger.warning("These binaries will likely cause issues with code-signing and hardened runtime!")

    def _write_warnings(self):
        """
        Write warnings about missing modules. Get them from the graph and use the graph to figure out who tried to
        import them.
        """
        def dependency_description(name, dep_info):
            if not dep_info or dep_info == 'direct':
                imptype = 0
            else:
                imptype = (dep_info.conditional + 2 * dep_info.function + 4 * dep_info.tryexcept)
            return '%s (%s)' % (name, IMPORT_TYPES[imptype])

        from PyInstaller.config import CONF
        miss_toc = self.graph.make_missing_toc()
        with open(CONF['warnfile'], 'w', encoding='utf-8') as wf:
            wf.write(WARNFILE_HEADER)
            for (n, p, status) in miss_toc:
                importers = self.graph.get_importers(n)
                print(
                    status,
                    'module named',
                    n,
                    '- imported by',
                    ', '.join(dependency_description(name, data) for name, data in importers),
                    file=wf
                )
        logger.info("Warnings written to %s", CONF['warnfile'])

    def _write_graph_debug(self):
        """
        Write a xref (in html) and with `--log-level DEBUG` a dot-drawing of the graph.
        """
        from PyInstaller.config import CONF
        with open(CONF['xref-file'], 'w', encoding='utf-8') as fh:
            self.graph.create_xref(fh)
            logger.info("Graph cross-reference written to %s", CONF['xref-file'])
        if logger.getEffectiveLevel() > logging.DEBUG:
            return
        # The `DOT language's <https://www.graphviz.org/doc/info/lang.html>`_ default character encoding (see the end
        # of the linked page) is UTF-8.
        with open(CONF['dot-file'], 'w', encoding='utf-8') as fh:
            self.graph.graphreport(fh)
            logger.info("Graph drawing written to %s", CONF['dot-file'])

    def exclude_system_libraries(self, list_of_exceptions=None):
        """
        This method may be optionally called from the spec file to exclude any system libraries from the list of
        binaries other than those containing the shell-style wildcards in list_of_exceptions. Those that match
        '*python*' or are stored under 'lib-dynload' are always treated as exceptions and not excluded.
        """

        self.binaries = [
            entry for entry in self.binaries if _should_include_system_binary(entry, list_of_exceptions or [])
        ]


class ExecutableBuilder:
    """
    Class that constructs the executable.
    """
    # TODO wrap the 'main' and 'build' function into this class.


def build(spec, distpath, workpath, clean_build):
    """
    Build the executable according to the created SPEC file.
    """
    from PyInstaller.config import CONF

    # Ensure starting tilde in distpath / workpath is expanded into user's home directory. This is to work around for
    # tilde not being expanded when using `--workpath=~/path/abc` instead of `--workpath ~/path/abc` (or when the path
    # argument is quoted). See https://github.com/pyinstaller/pyinstaller/issues/696
    distpath = os.path.abspath(os.path.expanduser(distpath))
    workpath = os.path.abspath(os.path.expanduser(workpath))

    CONF['spec'] = os.path.abspath(spec)
    CONF['specpath'], CONF['specnm'] = os.path.split(CONF['spec'])
    CONF['specnm'] = os.path.splitext(CONF['specnm'])[0]

    # Add 'specname' to workpath and distpath if they point to PyInstaller homepath.
    if os.path.dirname(distpath) == HOMEPATH:
        distpath = os.path.join(HOMEPATH, CONF['specnm'], os.path.basename(distpath))
    CONF['distpath'] = distpath
    if os.path.dirname(workpath) == HOMEPATH:
        workpath = os.path.join(HOMEPATH, CONF['specnm'], os.path.basename(workpath), CONF['specnm'])
    else:
        workpath = os.path.join(workpath, CONF['specnm'])
    CONF['workpath'] = workpath

    CONF['warnfile'] = os.path.join(workpath, 'warn-%s.txt' % CONF['specnm'])
    CONF['dot-file'] = os.path.join(workpath, 'graph-%s.dot' % CONF['specnm'])
    CONF['xref-file'] = os.path.join(workpath, 'xref-%s.html' % CONF['specnm'])

    CONF['code_cache'] = dict()

    # Clean PyInstaller cache (CONF['cachedir']) and temporary files (workpath) to be able start a clean build.
    if clean_build:
        logger.info('Removing temporary files and cleaning cache in %s', CONF['cachedir'])
        for pth in (CONF['cachedir'], workpath):
            if os.path.exists(pth):
                # Remove all files in 'pth'.
                for f in glob.glob(pth + '/*'):
                    # Remove dirs recursively.
                    if os.path.isdir(f):
                        shutil.rmtree(f)
                    else:
                        os.remove(f)

    # Create DISTPATH and workpath if they does not exist.
    for pth in (CONF['distpath'], CONF['workpath']):
        os.makedirs(pth, exist_ok=True)

    # Construct NAMESPACE for running the Python code from .SPEC file.
    # NOTE: Passing NAMESPACE allows to avoid having global variables in this module and makes isolated environment for
    #       running tests.
    # NOTE: Defining NAMESPACE allows to map any class to a apecific name for .SPEC.
    # FIXME: Some symbols might be missing. Add them if there are some failures.
    # TODO: What from this .spec API is deprecated and could be removed?
    spec_namespace = {
        # Set of global variables that can be used while processing .spec file. Some of them act as configuration
        # options.
        'DISTPATH': CONF['distpath'],
        'HOMEPATH': HOMEPATH,
        'SPEC': CONF['spec'],
        'specnm': CONF['specnm'],
        'SPECPATH': CONF['specpath'],
        'WARNFILE': CONF['warnfile'],
        'workpath': CONF['workpath'],
        # PyInstaller classes for .spec.
        'TOC': TOC,  # Kept for backward compatibility even though `TOC` class is deprecated.
        'Analysis': Analysis,
        'BUNDLE': BUNDLE,
        'COLLECT': COLLECT,
        'EXE': EXE,
        'MERGE': MERGE,
        'PYZ': PYZ,
        'Tree': Tree,
        'Splash': Splash,
        # Python modules available for .spec.
        'os': os,
    }

    # Execute the specfile. Read it as a binary file...
    try:
        with open(spec, 'rb') as f:
            # ... then let Python determine the encoding, since ``compile`` accepts byte strings.
            code = compile(f.read(), spec, 'exec')
    except FileNotFoundError:
        raise SystemExit(f'ERROR: Spec file "{spec}" not found!')
    exec(code, spec_namespace)

    logger.info("Build complete! The results are available in: %s", CONF['distpath'])


def __add_options(parser):
    parser.add_argument(
        "--distpath",
        metavar="DIR",
        default=DEFAULT_DISTPATH,
        help="Where to put the bundled app (default: ./dist)",
    )
    parser.add_argument(
        '--workpath',
        default=DEFAULT_WORKPATH,
        help="Where to put all the temporary work files, .log, .pyz and etc. (default: ./build)",
    )
    parser.add_argument(
        '-y',
        '--noconfirm',
        action="store_true",
        default=False,
        help="Replace output directory (default: %s) without asking for confirmation" %
        os.path.join('SPECPATH', 'dist', 'SPECNAME'),
    )
    parser.add_argument(
        '--upx-dir',
        default=None,
        help="Path to UPX utility (default: search the execution path)",
    )
    parser.add_argument(
        '--clean',
        dest='clean_build',
        action='store_true',
        default=False,
        help="Clean PyInstaller cache and remove temporary files before building.",
    )


def main(
    pyi_config,
    specfile,
    noconfirm=False,
    distpath=DEFAULT_DISTPATH,
    workpath=DEFAULT_WORKPATH,
    upx_dir=None,
    clean_build=False,
    **kw
):
    from PyInstaller.config import CONF
    CONF['noconfirm'] = noconfirm

    # If configuration dict is supplied - skip configuration step.
    if pyi_config is None:
        import PyInstaller.configure as configure
        CONF.update(configure.get_config(upx_dir=upx_dir))
    else:
        CONF.update(pyi_config)

    CONF['ui_admin'] = kw.get('ui_admin', False)
    CONF['ui_access'] = kw.get('ui_uiaccess', False)

    build(specfile, distpath, workpath, clean_build)
